"""
Test the expressions module

:author: Michel Bierlaire
:data: Wed Apr 29 17:47:53 2020

"""

import unittest

import numpy as np
import pandas as pd
from scipy.stats import norm

from biogeme import models
from biogeme.audit_tuple import AuditTuple
from biogeme.calculator import (
    CallableExpression,
    create_function_simple_expression,
    get_value_and_derivatives,
    get_value_c,
)
from biogeme.database import Database
from biogeme.distributions import normalpdf
from biogeme.exceptions import BiogemeError
from biogeme.expressions import (
    BelongsTo,
    Beta,
    BinaryMax,
    BinaryMin,
    ConditionalSum,
    ConditionalTermTuple,
    Derive,
    Draws,
    Elem,
    Expression,
    IntegrateNormal,
    LinearTermTuple,
    LinearUtility,
    LogLogit,
    MonteCarlo,
    MultipleSum,
    NormalCdf,
    Numeric,
    PanelLikelihoodTrajectory,
    RandomVariable,
    Variable,
    audit_expression,
    cos,
    exp,
    expression_to_value,
    get_dict_expressions,
    get_dict_values,
    is_numeric,
    list_of_all_betas_in_expression,
    list_of_fixed_betas_in_expression,
    list_of_free_betas_in_expression,
    list_of_random_variables_in_expression,
    list_of_variables_in_expression,
    log,
    logzero,
    sin,
    validate_and_convert,
)
from biogeme.expressions.collectors import list_of_draws_in_expression
from biogeme.expressions.minus import Minus
from biogeme.expressions_registry import ExpressionRegistry
from biogeme.function_output import (
    FunctionOutput,
    NamedFunctionOutput,
    convert_to_dict,
)
from biogeme.likelihood.negative_likelihood import NegativeLikelihood
from biogeme.tools import check_derivatives
from biogeme.tools.derivatives import CheckDerivativesResults
from test_data import getData

EPSILON: float = np.finfo(np.float64).eps


class test_expressions(unittest.TestCase):
    def setUp(self):
        self.myData = getData(2)
        self.Person = Variable('Person')
        self.Variable1 = Variable('Variable1')
        self.Variable2 = Variable('Variable2')
        self.Choice = Variable('Choice')
        self.Av1 = Variable('Av1')
        self.Av2 = Variable('Av2')
        self.Av3 = Variable('Av3')
        self.beta1 = Beta('beta1', 0.2, None, None, 0)
        self.beta2 = Beta('beta2', 0.4, None, None, 0)
        self.beta3 = Beta('beta3', 1, None, None, 1)
        self.beta4 = Beta('beta4', 0, None, None, 1)
        self.omega1 = RandomVariable('omega1')
        self.omega2 = RandomVariable('omega2')
        self.xi1 = Draws('xi1', 'NORMAL')
        self.xi2 = Draws('xi2', 'UNIF')
        self.xi3 = Draws('xi3', 'WRONGTYPE')
        self.addTypeEqualityFunc(pd.DataFrame, self.assertDataframeEqual)

    def test_create_function(self):
        beta1 = Beta('beta1', 0, None, None, 0)
        beta2 = Beta('beta2', 0, None, None, 0)
        quotient = beta1 / beta2
        the_function: CallableExpression = create_function_simple_expression(
            expression=quotient, numerically_safe=False
        )
        the_function_output: FunctionOutput = the_function(
            [2, 4], gradient=False, hessian=False, bhhh=False
        )
        self.assertEqual(2.0 / 4.0, the_function_output.function)

    def test_errors(self):
        with self.assertRaises(BiogemeError):
            _ = Numeric(1) / 'ert'

    #        with self.assertRaises(exceptions.BiogemeError):
    #            _ = 'ert' / Numeric(1)

    def assertDataframeEqual(self, a, b, msg):
        try:
            pd.testing.assert_frame_equal(a, b)
        except AssertionError as e:
            raise self.failureException(msg) from e

    def test_is_numeric(self):
        result = is_numeric(1)
        self.assertTrue(result)
        result = is_numeric(0.1)
        self.assertTrue(result)
        result = is_numeric(True)
        self.assertTrue(result)
        result = is_numeric(self)
        self.assertFalse(result)

    def test_add(self):
        result = self.Variable1 + self.Variable2
        self.assertEqual(str(result), '(Variable1 + Variable2)')
        self.assertTrue(result.children[0] is self.Variable1)
        self.assertTrue(result.children[1] is self.Variable2)

        result = self.Variable1 + 1
        self.assertEqual(str(result), '(Variable1 + `1.0`)')
        self.assertTrue(result.children[0] is self.Variable1)

        result = 1 + self.Variable1
        self.assertEqual(str(result), '(`1.0` + Variable1)')
        self.assertTrue(result.children[1] is self.Variable1)

        with self.assertRaises(BiogemeError):
            _ = self + self.Variable1

        with self.assertRaises(BiogemeError):
            _ = self.Variable1 + self

    def test_sub(self):
        result = self.Variable1 - self.Variable2
        self.assertEqual(str(result), '(Variable1 - Variable2)')
        self.assertTrue(result.children[0] is self.Variable1)
        self.assertTrue(result.children[1] is self.Variable2)

        result = self.Variable1 - 1
        self.assertEqual(str(result), '(Variable1 - `1.0`)')
        self.assertTrue(result.children[0] is self.Variable1)

        result = 1 - self.Variable1
        self.assertEqual(str(result), '(`1.0` - Variable1)')
        self.assertTrue(result.children[1] is self.Variable1)

        with self.assertRaises(BiogemeError):
            _ = self - self.Variable1

        with self.assertRaises(BiogemeError):
            _ = self.Variable1 - self

    def test_mul(self):
        result = self.Variable1 * self.Variable2
        self.assertEqual(str(result), '(Variable1 * Variable2)')
        self.assertTrue(result.children[0] is self.Variable1)
        self.assertTrue(result.children[1] is self.Variable2)

        result = self.Variable1 * 1
        self.assertEqual(str(result), '(Variable1 * `1.0`)')
        self.assertTrue(result.children[0] is self.Variable1)

        result = 1 * self.Variable1
        self.assertEqual(str(result), '(`1.0` * Variable1)')
        self.assertTrue(result.children[1] is self.Variable1)

        with self.assertRaises(BiogemeError):
            _ = self * self.Variable1

        with self.assertRaises(BiogemeError):
            _ = self.Variable1 * self

    def test_exp(self):
        argument = Beta('argument', 2, None, None, 0)
        the_exp = exp(argument)
        expression_function: CallableExpression = create_function_simple_expression(
            expression=the_exp, numerically_safe=False
        )
        the_function_output: FunctionOutput = expression_function(
            [10.0], gradient=False, hessian=False, bhhh=False
        )
        self.assertAlmostEqual(the_function_output.function, np.exp(10), 3)
        optimization_function: CallableExpression = create_function_simple_expression(
            expression=the_exp, numerically_safe=False
        )
        check_results: CheckDerivativesResults = check_derivatives(
            the_function=optimization_function, x=[1.0]
        )
        for a_grad, fd_grad in zip(
            check_results.analytical_gradient, check_results.finite_differences_gradient
        ):
            self.assertAlmostEqual(a_grad, fd_grad, delta=1.0e-4)
        for a_hess, fd_hess in zip(
            np.nditer(check_results.analytical_hessian),
            np.nditer(check_results.finite_differences_hessian),
        ):
            self.assertAlmostEqual(a_hess, fd_hess, delta=1.0e-4)

        # exp of a large number
        check_results = check_derivatives(
            the_function=optimization_function, x=[1800.0]
        )
        self.assertAlmostEqual(check_results.function, np.inf)

    def test_loglogit(self):
        V1 = Beta('V1', 2, None, None, 0)
        V2 = Beta('V2', 2, None, None, 0)
        V = {1: V1, 2: V2}
        av = {1: 1, 2: 1}
        the_logit = LogLogit(V, av, 1)
        expression_function = create_function_simple_expression(
            expression=the_logit, numerically_safe=False
        )
        the_function_output: FunctionOutput = expression_function(
            [10.0, 30], gradient=False, hessian=False, bhhh=False
        )
        self.assertAlmostEqual(the_function_output.function, -20.000000002061153, 3)
        optimization_function = create_function_simple_expression(
            expression=the_logit, numerically_safe=False
        )
        check_results = check_derivatives(
            the_function=optimization_function, x=[10, 11]
        )
        for a_grad, fd_grad in zip(
            check_results.analytical_gradient, check_results.finite_differences_gradient
        ):
            self.assertAlmostEqual(a_grad, fd_grad, delta=1.0e-4)
        for a_hess, fd_hess in zip(
            np.nditer(check_results.analytical_hessian),
            np.nditer(check_results.finite_differences_hessian),
        ):
            self.assertAlmostEqual(a_hess, fd_hess, delta=1.0e-4)

    def test_conjunction(self):
        zero = Numeric(0)
        non_zero = Numeric(12)
        other_non_zero = Numeric(-1)
        and_result = non_zero & other_non_zero
        should_be_true = get_value_and_derivatives(
            expression=and_result, numerically_safe=False
        )
        self.assertEqual(1, should_be_true.function)
        and_result_with_other_syntax = non_zero & other_non_zero
        should_be_true = get_value_and_derivatives(
            expression=and_result, numerically_safe=False
        )
        self.assertEqual(should_be_true.function, 1)
        other_result = zero & other_non_zero
        should_be_false = get_value_and_derivatives(
            expression=other_result, numerically_safe=False
        )
        self.assertEqual(should_be_false.function, 0)

    def test_disjunction(self):
        zero = Numeric(0)
        zero_two = Numeric(0)
        non_zero = Numeric(12)
        or_result = non_zero | zero
        should_be_true = get_value_and_derivatives(
            expression=or_result, numerically_safe=False
        )
        self.assertEqual(should_be_true.function, 1)
        other_result = zero | zero_two
        should_be_false = get_value_and_derivatives(
            expression=other_result, numerically_safe=False
        )
        self.assertEqual(should_be_false.function, 0)

    def test_loglogit_full_choice_set(self):
        V1 = Beta('V1', 2, None, None, 0)
        V2 = Beta('V2', 2, None, None, 0)
        V = {1: V1, 2: V2}
        the_logit = LogLogit(util=V, av=None, choice=1)
        expression_function: CallableExpression = create_function_simple_expression(
            expression=the_logit, numerically_safe=False
        )
        the_function_output: FunctionOutput = expression_function(
            [10.0, 30], gradient=False, hessian=False, bhhh=False
        )
        self.assertAlmostEqual(the_function_output.function, -20.000000002061153, 3)
        optimization_function = create_function_simple_expression(
            expression=the_logit, numerically_safe=False
        )
        check_results = check_derivatives(
            the_function=optimization_function, x=[10, 11]
        )
        for a_grad, fd_grad in zip(
            check_results.analytical_gradient, check_results.finite_differences_gradient
        ):
            self.assertAlmostEqual(a_grad, fd_grad, delta=1.0e-4)
        for a_hess, fd_hess in zip(
            np.nditer(check_results.analytical_hessian),
            np.nditer(check_results.finite_differences_hessian),
        ):
            self.assertAlmostEqual(a_hess, fd_hess, delta=1.0e-4)

    def test_normal_cdf(self):
        argument = Beta('argument', 2, None, None, 0)
        the_cdf = NormalCdf(argument)
        expression_function: CallableExpression = create_function_simple_expression(
            expression=the_cdf, numerically_safe=False
        )
        the_function_output: FunctionOutput = expression_function(
            [10.0], gradient=False, hessian=False, bhhh=False
        )
        self.assertAlmostEqual(the_function_output.function, norm.cdf(10), 3)
        the_function_output: FunctionOutput = expression_function(
            [-0.01], gradient=False, hessian=False, bhhh=False
        )
        self.assertAlmostEqual(the_function_output.function, norm.cdf(-0.01), 3)

        optimization_function = create_function_simple_expression(
            expression=the_cdf, numerically_safe=False
        )
        check_results = check_derivatives(the_function=optimization_function, x=[1.0])
        for a_grad, fd_grad in zip(
            check_results.analytical_gradient, check_results.finite_differences_gradient
        ):
            self.assertAlmostEqual(a_grad, fd_grad, delta=1.0e-4)
        for a_hess, fd_hess in zip(
            np.nditer(check_results.analytical_hessian),
            np.nditer(check_results.finite_differences_hessian),
        ):
            self.assertAlmostEqual(a_hess, fd_hess, delta=1.0e-4)

    def test_power_constant(self):
        x = Beta('x', 2, None, None, 0)
        x_square = x * x
        the_power = x_square**2
        expression_function: CallableExpression = create_function_simple_expression(
            expression=the_power, numerically_safe=False
        )
        the_function_output: FunctionOutput = expression_function(
            [2.0], gradient=False, hessian=False, bhhh=False
        )
        self.assertAlmostEqual(the_function_output.function, 16, 3)
        check_results = check_derivatives(the_function=expression_function, x=[1.0])
        for a_grad, fd_grad in zip(
            check_results.analytical_gradient, check_results.finite_differences_gradient
        ):
            self.assertAlmostEqual(a_grad, fd_grad, delta=1.0e-4)
        for a_hess, fd_hess in zip(
            np.nditer(check_results.analytical_hessian),
            np.nditer(check_results.finite_differences_hessian),
        ):
            self.assertAlmostEqual(a_hess, fd_hess, delta=1.0e-4)

        # Test a negative argument and an integer
        check_results = check_derivatives(the_function=expression_function, x=[-1.0])
        for a_grad, fd_grad in zip(
            check_results.analytical_gradient, check_results.finite_differences_gradient
        ):
            self.assertAlmostEqual(a_grad, fd_grad, delta=1.0e-4)
        for a_hess, fd_hess in zip(
            np.nditer(check_results.analytical_hessian),
            np.nditer(check_results.finite_differences_hessian),
        ):
            self.assertAlmostEqual(a_hess, fd_hess, delta=1.0e-4)

    def test_power_constant_neg_integer(self):
        x = Beta('x', 2, None, None, 0)
        # Test a negative argument and an integer
        the_power = x**-2
        expression_function: CallableExpression = create_function_simple_expression(
            expression=the_power, numerically_safe=False
        )
        the_function_output: FunctionOutput = expression_function(
            [-2.0], gradient=False, hessian=False, bhhh=False
        )
        self.assertAlmostEqual(the_function_output.function, 0.25, 3)
        check_results: CheckDerivativesResults = check_derivatives(
            the_function=expression_function, x=[-1.0]
        )
        for a_grad, fd_grad in zip(
            check_results.analytical_gradient, check_results.finite_differences_gradient
        ):
            self.assertAlmostEqual(a_grad, fd_grad, delta=1.0e-4)
        for a_hess, fd_hess in zip(
            np.nditer(check_results.analytical_hessian),
            np.nditer(check_results.finite_differences_hessian),
        ):
            self.assertAlmostEqual(a_hess, fd_hess, delta=1.0e-4)

    def test_power_constant_epsilon(self):
        x = Beta('x', 2, None, None, 0)
        # Test a negative argument and an integer
        the_power = x**2
        small_x = 1.0e-34
        expression_function: CallableExpression = create_function_simple_expression(
            expression=the_power, numerically_safe=False
        )
        the_function_output: FunctionOutput = expression_function(
            [small_x], gradient=True, hessian=True, bhhh=True
        )
        self.assertAlmostEqual(the_function_output.function, 0)
        self.assertAlmostEqual(the_function_output.gradient[0], 0)
        self.assertAlmostEqual(the_function_output.hessian[0][0], 2)

    def test_power_constant_neg_non_integer(self):
        # -2 ** -2.5: this must raise an exception
        x = Beta('x', 2, None, None, 0)
        # Test a negative argument and an integer
        the_power = x**-2.5
        expression_function = create_function_simple_expression(
            expression=the_power, numerically_safe=False
        )

        negative_number = expression_function(
            [-2.0], gradient=False, hessian=False, bhhh=False
        )
        self.assertTrue(np.isnan(negative_number.function), 'The value is not NaN')

    def test_power(self):
        x = Beta('x', 2, None, None, 0)
        y = Beta('y', 2, None, None, 0)

        other_power = (x * x) ** (y + y)
        expression_function: CallableExpression = create_function_simple_expression(
            expression=other_power, numerically_safe=False
        )
        the_function_output: FunctionOutput = expression_function(
            [2, 1], gradient=False, hessian=False, bhhh=False
        )
        self.assertAlmostEqual(the_function_output.function, 16, 3)
        check_results = check_derivatives(
            the_function=expression_function, x=[1.0, 1.0]
        )
        for a_grad, fd_grad in zip(
            check_results.analytical_gradient, check_results.finite_differences_gradient
        ):
            self.assertAlmostEqual(a_grad, fd_grad, delta=1.0e-4)
        for a_hess, fd_hess in zip(
            np.nditer(check_results.analytical_hessian),
            np.nditer(check_results.finite_differences_hessian),
        ):
            self.assertAlmostEqual(a_hess, fd_hess, delta=1.0e-4)

    def test_power_small_base(self):
        x = Beta('x', 2, None, None, 0)
        y = Beta('y', 2, None, None, 0)

        other_power = (x * x) ** (y + y)
        expression_function: CallableExpression = create_function_simple_expression(
            expression=other_power, numerically_safe=False
        )
        the_function_output: FunctionOutput = expression_function(
            [2, 1], gradient=False, hessian=False, bhhh=False
        )
        self.assertAlmostEqual(the_function_output.function, 16, 3)
        check_results = check_derivatives(
            the_function=expression_function, x=[1.0, 1.0]
        )
        for a_grad, fd_grad in zip(
            check_results.analytical_gradient, check_results.finite_differences_gradient
        ):
            self.assertAlmostEqual(a_grad, fd_grad, delta=1.0e-4)
        for a_hess, fd_hess in zip(
            np.nditer(check_results.analytical_hessian),
            np.nditer(check_results.finite_differences_hessian),
        ):
            self.assertAlmostEqual(a_hess, fd_hess, delta=1.0e-4)

    def test_power_neg_non_integer(self):
        # -2 ** -2.5: this must raise an exception
        x = Beta('x', 2, None, None, 0)
        exponent = Beta('exponent', -2.5, None, None, 0)
        # Test a negative argument and an integer
        the_power = x**exponent
        expression_function = create_function_simple_expression(
            expression=the_power, numerically_safe=False
        )

        negative_number = expression_function(
            [-2.0], gradient=False, hessian=False, bhhh=False
        )
        self.assertTrue(np.isnan(negative_number.function), 'The value is not NaN')

    def test_power_using_constant(self):
        x = Beta('x', 2, None, None, 0)
        y = Numeric(1) * Numeric(1)

        the_power = (x * x) ** (y + y)
        optimization_function = NegativeLikelihood(
            dimension=2,
            loglikelihood=create_function_simple_expression(
                expression=the_power, numerically_safe=False
            ),
        )
        check_results = optimization_function.check_derivatives(x=[1.0])
        for a_grad, fd_grad in zip(
            check_results.analytical.gradient, check_results.finite_differences.gradient
        ):
            self.assertAlmostEqual(a_grad, fd_grad, places=2)
        for a_hess, fd_hess in zip(
            np.nditer(check_results.analytical.hessian),
            np.nditer(check_results.finite_differences.hessian),
        ):
            self.assertAlmostEqual(a_hess, fd_hess, delta=1.0e-4)

    def test_complex_expression(self):
        df1 = pd.DataFrame(
            {
                'Person': [1, 1, 1, 2, 2],
                'Exclude': [0, 0, 1, 0, 1],
                'Variable1': [1, 2, 3, 4, 5],
                'Variable2': [10, 20, 30, 40, 50],
                'Choice': [1, 2, 3, 1, 2],
                'Av1': [0, 1, 1, 1, 1],
                'Av2': [1, 1, 1, 1, 1],
                'Av3': [0, 1, 1, 1, 1],
            }
        )
        data = Database(f'test', df1)
        variable1 = Variable('Variable1')
        variable2 = Variable('Variable2')
        beta1 = Beta('beta1', -1.0, None, None, 0)
        beta2 = Beta('beta2', 2.0, None, None, 0)
        likelihood = -((beta1 * variable1) ** 2) - (beta2 * variable2) ** 2
        optimization_function = NegativeLikelihood(
            dimension=2,
            loglikelihood=create_function_simple_expression(
                expression=likelihood, database=data, numerically_safe=False
            ),
        )
        check_results = optimization_function.check_derivatives(x=[-1.0, -1.0])
        for a_grad, fd_grad in zip(
            check_results.analytical.gradient, check_results.finite_differences.gradient
        ):
            self.assertAlmostEqual(a_grad, fd_grad, places=2)
        for a_hess, fd_hess in zip(
            np.nditer(check_results.analytical.hessian),
            np.nditer(check_results.finite_differences.hessian),
        ):
            self.assertAlmostEqual(a_hess, fd_hess, delta=1.0e-4)

    def test_log(self):
        argument = Beta('argument', 2, None, None, 0)
        the_log = log(argument)
        expression_function = create_function_simple_expression(
            expression=the_log, numerically_safe=False
        )
        the_function_output: FunctionOutput = expression_function(
            [10.0], gradient=False, hessian=False, bhhh=False
        )
        self.assertAlmostEqual(the_function_output.function, np.log(10), 3)
        optimization_function = NegativeLikelihood(
            dimension=1,
            loglikelihood=create_function_simple_expression(
                expression=the_log, numerically_safe=False
            ),
        )
        check_results = optimization_function.check_derivatives(x=[1.0])
        for a_grad, fd_grad in zip(
            check_results.analytical.gradient, check_results.finite_differences.gradient
        ):
            self.assertAlmostEqual(a_grad, fd_grad, places=2)
        for a_hess, fd_hess in zip(
            np.nditer(check_results.analytical.hessian),
            np.nditer(check_results.finite_differences.hessian),
        ):
            self.assertAlmostEqual(a_hess, fd_hess, delta=1.0e-4)

        # Log of a negative number
        log_negative = expression_function(
            [-10.0], gradient=False, hessian=False, bhhh=False
        )
        self.assertTrue(np.isnan(log_negative.function), 'The value is not NaN')

        # Log of zero
        log_of_zero = expression_function(
            [0.0], gradient=False, hessian=False, bhhh=False
        )
        self.assertTrue(np.isinf(log_of_zero.function), 'The value is not infinity')

        # Log of a number close to zero
        log_small_number = expression_function(
            [EPSILON], gradient=False, hessian=False, bhhh=False
        )
        self.assertEqual(
            np.log(EPSILON),
            log_small_number.function,
        )

        # Log of a number very close to zero
        log_small_number = expression_function(
            [EPSILON / 2], gradient=False, hessian=False, bhhh=False
        )
        self.assertEqual(np.log(EPSILON / 2), log_small_number.function)

    def test_logzero(self):
        argument = Beta('argument', 2, None, None, 0)
        the_log = logzero(argument)
        expression_function = create_function_simple_expression(
            expression=the_log, numerically_safe=False
        )
        the_function_output: FunctionOutput = expression_function(
            [10.0], gradient=False, hessian=False, bhhh=False
        )
        self.assertAlmostEqual(the_function_output.function, np.log(10), 3)
        optimization_function = NegativeLikelihood(
            dimension=1,
            loglikelihood=expression_function,
        )
        check_results = optimization_function.check_derivatives(x=[1.0])
        for a_grad, fd_grad in zip(
            check_results.analytical.gradient, check_results.finite_differences.gradient
        ):
            self.assertAlmostEqual(a_grad, fd_grad, places=3)
        for a_hess, fd_hess in zip(
            np.nditer(check_results.analytical.hessian),
            np.nditer(check_results.finite_differences.hessian),
        ):
            self.assertAlmostEqual(a_hess, fd_hess, delta=1.0e-4)

        # Log of a negative number
        log_negative = expression_function(
            [-10.0], gradient=False, hessian=False, bhhh=False
        )
        self.assertTrue(np.isnan(log_negative.function), 'The value is not NaN')

        # Log of zero
        log_of_zero = expression_function(
            [0.0], gradient=False, hessian=False, bhhh=False
        )
        self.assertEqual(
            0,
            log_of_zero.function,
        )

        # Log of a number close to zero
        log_small_number = expression_function(
            [EPSILON], gradient=False, hessian=False, bhhh=False
        )
        self.assertEqual(
            np.log(EPSILON),
            log_small_number.function,
        )

        # Log of a number very close to zero
        log_small_number = expression_function(
            [EPSILON / 2], gradient=False, hessian=False, bhhh=False
        )
        self.assertEqual(np.log(EPSILON / 2), log_small_number.function)

    def test_div(self):
        result = self.Variable1 / self.Variable2
        self.assertEqual(str(result), '(Variable1 / Variable2)')
        self.assertTrue(result.children[0] is self.Variable1)
        self.assertTrue(result.children[1] is self.Variable2)

        # Numbering of the variables is by alphabetical order.
        numerator = Beta('a_numerator', 2, None, None, 0)
        denominator = Beta('b_denominator', 2, None, None, 0)
        the_ratio = numerator / denominator
        expression_function = create_function_simple_expression(
            expression=the_ratio, numerically_safe=False
        )
        the_function_output: FunctionOutput = expression_function(
            [2.0, 2.0], gradient=False, hessian=False, bhhh=False
        )
        self.assertAlmostEqual(the_function_output.function, 1, 3)
        optimization_function = NegativeLikelihood(
            dimension=2,
            loglikelihood=expression_function,
        )
        optimization_function.x = [2.0, 4.0]
        check_results = optimization_function.check_derivatives(x=[2.0, 4.0])
        for a_grad, fd_grad in zip(
            check_results.analytical.gradient, check_results.finite_differences.gradient
        ):
            self.assertAlmostEqual(a_grad, fd_grad, delta=1.0e-4)
        for a_hess, fd_hess in zip(
            np.nditer(check_results.analytical.hessian),
            np.nditer(check_results.finite_differences.hessian),
        ):
            self.assertAlmostEqual(a_hess, fd_hess, delta=1.0e-4)
        # Special case where numerator is 0
        the_function_output: FunctionOutput = expression_function(
            [0.0, 2.0], gradient=False, hessian=False, bhhh=False
        )
        self.assertAlmostEqual(the_function_output.function, 0, 3)
        check_results = optimization_function.check_derivatives(x=[0.0, 2.0])
        for a_grad, fd_grad in zip(
            check_results.analytical.gradient, check_results.finite_differences.gradient
        ):
            self.assertAlmostEqual(a_grad, fd_grad, delta=1.0e-4)
        for a_hess, fd_hess in zip(
            np.nditer(check_results.analytical.hessian),
            np.nditer(check_results.finite_differences.hessian),
        ):
            self.assertAlmostEqual(a_hess, fd_hess, delta=1.0e-4)

        # Special case where denominator is 1
        the_function_output: FunctionOutput = expression_function(
            [2.0, 1.0], gradient=False, hessian=False, bhhh=False
        )
        self.assertAlmostEqual(the_function_output.function, 2, 3)
        check_results = optimization_function.check_derivatives(x=[2.0, 1.0])
        for a_grad, fd_grad in zip(
            check_results.analytical.gradient, check_results.finite_differences.gradient
        ):
            self.assertAlmostEqual(a_grad, fd_grad, places=2)
        for a_hess, fd_hess in zip(
            np.nditer(check_results.analytical.hessian),
            np.nditer(check_results.finite_differences.hessian),
        ):
            self.assertAlmostEqual(a_hess, fd_hess, delta=1.0e-4)

        # Special case where denominator is close to zero
        the_function_output: FunctionOutput = expression_function(
            [2.0, 1.0e-18], gradient=False, hessian=False, bhhh=False
        )
        self.assertAlmostEqual(the_function_output.function, 2e18, delta=1e12)
        # Here, no way to verify the derivatives with finite difference, due to the numerical difficulties.

        # Special case where denominator is zero
        the_function_output: FunctionOutput = expression_function(
            [2, 0], gradient=False, hessian=False, bhhh=False
        )
        self.assertTrue(
            np.isinf(the_function_output.function), 'The value is not infinity'
        )
        # Here, no way to verify the derivatives with finite difference, due to the numerical difficulties.

        # Special case where denominator is close to zero
        the_function_output: FunctionOutput = expression_function(
            [2.0, -1.0e-18], gradient=False, hessian=False, bhhh=False
        )
        self.assertAlmostEqual(the_function_output.function, -2e18, delta=1e12)
        # Here, no way to verify the derivatives with finite difference, due to the numerical difficulties.

        result = self.Variable1 / 1
        self.assertEqual(str(result), '(Variable1 / `1.0`)')
        self.assertTrue(result.children[0] is self.Variable1)

        result = 1 / self.Variable1
        self.assertEqual(str(result), '(`1.0` / Variable1)')
        self.assertTrue(result.children[1] is self.Variable1)

        with self.assertRaises(BiogemeError):
            _ = self / self.Variable1

        with self.assertRaises(BiogemeError):
            _ = self.Variable1 / self

    def test_neg(self):
        result = -self.Variable1
        self.assertEqual(str(result), '(-Variable1)')
        self.assertTrue(result.children[0] is self.Variable1)

    def test_pow(self):
        result = self.Variable1**self.Variable2
        self.assertEqual(str(result), '(Variable1 ** Variable2)')
        self.assertTrue(result.children[0] is self.Variable1)
        self.assertTrue(result.children[1] is self.Variable2)

        result = self.Variable1**1
        self.assertEqual(str(result), 'Variable1**1.0')
        self.assertTrue(result.children[0] is self.Variable1)

        result = 1**self.Variable1
        self.assertEqual(str(result), '(`1.0` ** Variable1)')
        self.assertTrue(result.children[1] is self.Variable1)

        result = Numeric(3) ** Numeric(2)
        nine = result.get_value()
        self.assertEqual(9, nine)

        with self.assertRaises(BiogemeError):
            _ = self**self.Variable1

        with self.assertRaises(BiogemeError):
            _ = self.Variable1**self

    def test_and_1(self):
        result = self.Variable1 & self.Variable2
        self.assertEqual(str(result), '(Variable1 and Variable2)')
        self.assertTrue(result.children[0] is self.Variable1)
        self.assertTrue(result.children[1] is self.Variable2)

        result = self.Variable1 & 1
        self.assertEqual(str(result), '(Variable1 and `1.0`)')
        self.assertTrue(result.children[0] is self.Variable1)

        result = 1 & self.Variable1
        self.assertEqual(str(result), '(`1.0` and Variable1)')
        self.assertTrue(result.children[1] is self.Variable1)

        with self.assertRaises(BiogemeError):
            _ = self & self.Variable1

        with self.assertRaises(BiogemeError):
            _ = self.Variable1 & self

    def test_or(self):
        result = self.Variable1 | self.Variable2
        self.assertEqual(str(result), '(Variable1 or Variable2)')
        self.assertTrue(result.children[0] is self.Variable1)
        self.assertTrue(result.children[1] is self.Variable2)

        result = self.Variable1 | 1
        self.assertEqual(str(result), '(Variable1 or `1.0`)')
        self.assertTrue(result.children[0] is self.Variable1)

        result = 1 | self.Variable1
        self.assertEqual(str(result), '(`1.0` or Variable1)')
        self.assertTrue(result.children[1] is self.Variable1)

        with self.assertRaises(BiogemeError):
            _ = self | self.Variable1

        with self.assertRaises(BiogemeError):
            _ = self.Variable1 | self

    def test_eq(self):
        result = self.Variable1 == self.Variable2
        self.assertEqual(str(result), '(Variable1 == Variable2)')
        self.assertTrue(result.children[0] is self.Variable1)
        self.assertTrue(result.children[1] is self.Variable2)

        result = self.Variable1 == 1
        self.assertEqual(str(result), '(Variable1 == `1.0`)')
        self.assertTrue(result.children[0] is self.Variable1)

        result = 1 == self.Variable1
        self.assertEqual(str(result), '(Variable1 == `1.0`)')
        self.assertTrue(result.children[0] is self.Variable1)

        with self.assertRaises(BiogemeError):
            _ = self == self.Variable1

        with self.assertRaises(BiogemeError):
            _ = self.Variable1 == self

    def test_neq(self):
        result = self.Variable1 != self.Variable2
        self.assertEqual(str(result), '(Variable1 != Variable2)')
        self.assertTrue(result.children[0] is self.Variable1)
        self.assertTrue(result.children[1] is self.Variable2)

        result = self.Variable1 != 1
        self.assertEqual(str(result), '(Variable1 != `1.0`)')
        self.assertTrue(result.children[0] is self.Variable1)

        result = 1 != self.Variable1
        self.assertEqual(str(result), '(Variable1 != `1.0`)')
        self.assertTrue(result.children[0] is self.Variable1)

        with self.assertRaises(BiogemeError):
            _ = self != self.Variable1

        with self.assertRaises(BiogemeError):
            _ = self.Variable1 != self

    def test_le(self):
        result = self.Variable1 <= self.Variable2
        self.assertEqual(str(result), '(Variable1 <= Variable2)')
        self.assertTrue(result.children[0] is self.Variable1)
        self.assertTrue(result.children[1] is self.Variable2)

        result = self.Variable1 <= 1
        self.assertEqual(str(result), '(Variable1 <= `1.0`)')
        self.assertTrue(result.children[0] is self.Variable1)

        result = 1 <= self.Variable1
        self.assertEqual(str(result), '(Variable1 >= `1.0`)')
        self.assertTrue(result.children[0] is self.Variable1)

        with self.assertRaises(BiogemeError):
            _ = self <= self.Variable1

        with self.assertRaises(BiogemeError):
            _ = self.Variable1 <= self

    def test_ge(self):
        result = self.Variable1 >= self.Variable2
        self.assertEqual(str(result), '(Variable1 >= Variable2)')
        self.assertTrue(result.children[0] is self.Variable1)
        self.assertTrue(result.children[1] is self.Variable2)

        result = self.Variable1 >= 1
        self.assertEqual(str(result), '(Variable1 >= `1.0`)')
        self.assertTrue(result.children[0] is self.Variable1)

        result = 1 >= self.Variable1
        self.assertEqual(str(result), '(Variable1 <= `1.0`)')
        self.assertTrue(result.children[0] is self.Variable1)

        with self.assertRaises(BiogemeError):
            _ = self >= self.Variable1

        with self.assertRaises(BiogemeError):
            _ = self.Variable1 >= self

    def test_lt(self):
        result = self.Variable1 < self.Variable2
        self.assertEqual(str(result), '(Variable1 < Variable2)')
        self.assertTrue(result.children[0] is self.Variable1)
        self.assertTrue(result.children[1] is self.Variable2)

        result = self.Variable1 < 1
        self.assertEqual(str(result), '(Variable1 < `1.0`)')
        self.assertTrue(result.children[0] is self.Variable1)

        result = 1 < self.Variable1
        self.assertEqual(str(result), '(Variable1 > `1.0`)')
        self.assertTrue(result.children[0] is self.Variable1)

        with self.assertRaises(BiogemeError):
            _ = self < self.Variable1

        with self.assertRaises(BiogemeError):
            _ = self.Variable1 < self

    def test_gt(self):
        result = self.Variable1 > self.Variable2
        self.assertEqual(str(result), '(Variable1 > Variable2)')
        self.assertTrue(result.children[0] is self.Variable1)
        self.assertTrue(result.children[1] is self.Variable2)

        result = self.Variable1 > 1
        self.assertEqual(str(result), '(Variable1 > `1.0`)')
        self.assertTrue(result.children[0] is self.Variable1)

        result = 1 > self.Variable1
        self.assertEqual(str(result), '(Variable1 < `1.0`)')
        self.assertTrue(result.children[0] is self.Variable1)

        with self.assertRaises(BiogemeError):
            _ = self > self.Variable1

        with self.assertRaises(BiogemeError):
            _ = self.Variable1 > self

    def test_get_value_c(self):
        result = get_value_c(
            expression=self.Variable1, database=self.myData, numerically_safe=False
        )
        np.testing.assert_equal(result, [10, 20, 30, 40, 50])

    def test_DefineVariable(self):
        _ = self.myData.define_variable('newvar_b', self.Variable1 + self.Variable2)
        cols = self.myData.dataframe.columns
        added = 'newvar_b' in cols
        self.assertTrue(added)
        self.myData.dataframe['newvar_p'] = (
            self.myData.dataframe['Variable1'] + self.myData.dataframe['Variable2']
        )
        pd.testing.assert_series_equal(
            self.myData.dataframe['newvar_b'],
            self.myData.dataframe['newvar_p'],
            check_dtype=False,
            check_names=False,
        )

    def test_expr1(self):
        self.beta1.status = 1
        self.beta2.status = 1
        expr1 = 2 * self.beta1 - exp(-self.beta2) / (
            self.beta2 * (self.beta3 >= self.beta4)
            + self.beta1 * (self.beta3 < self.beta4)
        )
        self.assertAlmostEqual(expr1.get_value(), -1.275800115089098, 5)
        res = get_value_c(expression=expr1, numerically_safe=False)
        self.assertAlmostEqual(res, -1.275800115089098, 5)

    def test_expr1_newvalues(self):
        self.beta1.status = 1
        self.beta2.status = 1
        expr1 = 2 * self.beta1 - exp(-self.beta2) / (
            self.beta2 * (self.beta3 >= self.beta4)
            + self.beta1 * (self.beta3 < self.beta4)
        )
        new_values = {'beta1': 1, 'beta2': 2, 'beta3': 3, 'beta4': 2}
        expr1.change_init_values(new_values)
        self.assertAlmostEqual(expr1.get_value(), 1.9323323583816936, 5)
        res = get_value_c(expression=expr1, numerically_safe=False)
        self.assertAlmostEqual(res, 1.9323323583816936, 5)

    def test_expr1_derivatives(self):
        expr1 = 2 * self.beta1 - exp(-self.beta2) / (
            self.beta2 * (self.beta3 >= self.beta4)
            + self.beta1 * (self.beta3 < self.beta4)
        )
        newvalues = {'beta1': 1, 'beta2': 2, 'beta3': 3, 'beta4': 2}
        expr1.change_init_values(newvalues)
        the_function_output: FunctionOutput = get_value_and_derivatives(
            expression=expr1, numerically_safe=False
        )
        self.assertAlmostEqual(the_function_output.function, 1.9323323583816936, 5)
        g_ok = [2.0, 0.10150146242745953]
        h0_ok = [0.0, 0.0]
        h1_ok = [0.0, -0.16916910404576588]
        bhhh0_ok = [4, 0.20300292]
        bhhh1_ok = [0.20300292, 0.01030255]
        for check_left, check_right in zip(the_function_output.gradient, g_ok):
            self.assertAlmostEqual(check_left, check_right, 5)
        for check_left, check_right in zip(the_function_output.hessian[0], h0_ok):
            self.assertAlmostEqual(check_left, check_right, 5)
        for check_left, check_right in zip(the_function_output.hessian[1], h1_ok):
            self.assertAlmostEqual(check_left, check_right, 5)
        for check_left, check_right in zip(the_function_output.bhhh[0], bhhh0_ok):
            self.assertAlmostEqual(check_left, check_right, 5)
        for check_left, check_right in zip(the_function_output.bhhh[1], bhhh1_ok):
            self.assertAlmostEqual(check_left, check_right, 5)

    def test_expr1_named_derivatives(self):
        expr1 = 2 * self.beta1 - exp(-self.beta2) / (
            self.beta2 * (self.beta3 >= self.beta4)
            + self.beta1 * (self.beta3 < self.beta4)
        )
        new_values = {'beta1': 1, 'beta2': 2, 'beta3': 3, 'beta4': 2}
        expr1.change_init_values(new_values)
        the_function_output: NamedFunctionOutput = get_value_and_derivatives(
            expression=expr1, named_results=True, numerically_safe=False
        )
        self.assertAlmostEqual(the_function_output.function, 1.9323323583816936, 5)
        g_ok = {'beta1': 2.0, 'beta2': 0.10150146242745953}
        self.assertDictEqual(the_function_output.gradient, g_ok)
        h_ok = {
            'beta1': {'beta1': 0.0, 'beta2': 0.0},
            'beta2': {'beta1': 0.0, 'beta2': -0.16916910404576588},
        }
        self.assertDictEqual(the_function_output.hessian, h_ok)
        bhhh_ok = {
            'beta1': {'beta1': 4, 'beta2': 0.20300292485491905},
            'beta2': {'beta1': 0.20300292485491905, 'beta2': 0.010302546874912978},
        }
        self.assertDictEqual(the_function_output.bhhh, bhhh_ok)

    def test_expr1_gradient(self):
        expr1 = 2 * self.beta1 - exp(-self.beta2) / (
            self.beta2 * (self.beta3 >= self.beta4)
            + self.beta1 * (self.beta3 < self.beta4)
        )
        the_function_output: FunctionOutput = get_value_and_derivatives(
            expression=expr1,
            gradient=True,
            hessian=False,
            bhhh=False,
            numerically_safe=False,
        )
        self.assertAlmostEqual(the_function_output.function, -1.275800115089098, 5)
        g_ok = [2.0, 5.865300402811844]
        for check_left, check_right in zip(the_function_output.gradient, g_ok):
            self.assertAlmostEqual(check_left, check_right, 5)
        self.assertIsNone(the_function_output.hessian)
        self.assertIsNone(the_function_output.bhhh)

    def test_expr1_function(self):
        expr1 = 2 * self.beta1 - exp(-self.beta2) / (
            self.beta2 * (self.beta3 >= self.beta4)
            + self.beta1 * (self.beta3 < self.beta4)
        )
        the_function = create_function_simple_expression(
            expression=expr1, numerically_safe=False
        )
        the_function_output: FunctionOutput = the_function(
            [1, 2], gradient=True, hessian=True, bhhh=False
        )
        self.assertAlmostEqual(the_function_output.function, 1.9323323583816936, 5)
        g_ok = [2.0, 0.10150146242745953]
        h0_ok = [0.0, 0.0]
        h1_ok = [0.0, -0.16916910404576588]
        for check_left, check_right in zip(the_function_output.gradient, g_ok):
            self.assertAlmostEqual(check_left, check_right, 5)
        for check_left, check_right in zip(the_function_output.hessian[0], h0_ok):
            self.assertAlmostEqual(check_left, check_right, 5)
        for check_left, check_right in zip(the_function_output.hessian[1], h1_ok):
            self.assertAlmostEqual(check_left, check_right, 5)

        the_function_output: FunctionOutput = the_function(
            [10, -2], gradient=True, hessian=True, bhhh=False
        )
        self.assertAlmostEqual(the_function_output.function, 23.694528049465326, 5)
        g_ok = [2, -1.84726402]
        h0_ok = [0.0, 0.0]
        h1_ok = [0.0, 1.84726402]
        for check_left, check_right in zip(the_function_output.gradient, g_ok):
            self.assertAlmostEqual(check_left, check_right, 5)
        for check_left, check_right in zip(the_function_output.hessian[0], h0_ok):
            self.assertAlmostEqual(check_left, check_right, 5)
        for check_left, check_right in zip(the_function_output.hessian[1], h1_ok):
            self.assertAlmostEqual(check_left, check_right, 5)

    def test_expr1_database_agg(self):
        expr1 = 2 * self.beta1 - exp(-self.beta2) / (
            self.beta2 * (self.beta3 >= self.beta4)
            + self.beta1 * (self.beta3 < self.beta4)
        )
        the_function_output: FunctionOutput = get_value_and_derivatives(
            expression=expr1, database=self.myData, numerically_safe=False
        )
        f_ok = -6.37900057544549
        g_ok = [10.0, 29.32650201405922]
        h0_ok = [0.0, 0.0]
        h1_ok = [0.0, -155.01151064574157]
        b0_ok = [20.0, 58.65300403]
        b1_ok = [58.65300403, 172.00874408]
        self.assertAlmostEqual(the_function_output.function, f_ok, 5)
        for check_left, check_right in zip(the_function_output.gradient, g_ok):
            self.assertAlmostEqual(check_left, check_right, 5)
        for check_left, check_right in zip(the_function_output.hessian[0], h0_ok):
            self.assertAlmostEqual(check_left, check_right, 5)
        for check_left, check_right in zip(the_function_output.hessian[1], h1_ok):
            self.assertAlmostEqual(check_left, check_right, 5)
        for check_left, check_right in zip(the_function_output.bhhh[0], b0_ok):
            self.assertAlmostEqual(check_left, check_right, 5)
        for check_left, check_right in zip(the_function_output.bhhh[1], b1_ok):
            self.assertAlmostEqual(check_left, check_right, 5)

    def test_expr2(self):
        expr2 = 2 * self.beta1 * self.Variable1 - exp(-self.beta2 * self.Variable2) / (
            self.beta2 * (self.beta3 >= self.beta4)
            + self.beta1 * (self.beta3 < self.beta4)
        )
        with self.assertRaises(BiogemeError):
            expr2.get_value()
        res = list(
            get_value_c(expression=expr2, database=self.myData, numerically_safe=False)
        )
        self.assertListEqual(res, [4.0, 8.0, 12.0, 16.0, 20.0])

    def test_list_of_variables(self):
        expr2 = 2 * self.beta1 * self.Variable1 - exp(-self.beta2 * self.Variable2) / (
            self.beta3 * (self.beta2 >= self.beta1)
        )
        the_list = list_of_variables_in_expression(expr2)
        the_expected_list = [self.Variable1, self.Variable2]
        self.assertCountEqual(
            [str(x) for x in the_list],
            [str(x) for x in the_expected_list],
        )

    def test_list_of_betas(self):
        expr2 = 2 * self.beta1 * self.Variable1 - exp(-self.beta2 * self.Variable2) / (
            self.beta3 * (self.beta2 >= self.beta1)
        )
        the_list = list_of_all_betas_in_expression(expr2)
        the_expected_list = [self.beta1, self.beta2, self.beta3]
        self.assertCountEqual(
            {str(x) for x in the_list},
            {str(x) for x in the_expected_list},
        )
        the_list_free = list_of_free_betas_in_expression(expr2)
        the_expected_list_free = [self.beta1, self.beta2]
        self.assertCountEqual(
            {str(x) for x in the_list_free},
            {str(x) for x in the_expected_list_free},
        )
        the_list_fixed = list_of_fixed_betas_in_expression(expr2)
        the_expected_list_fixed = [self.beta3]
        self.assertCountEqual(
            {str(x) for x in the_list_fixed},
            {str(x) for x in the_expected_list_fixed},
        )
        expr_numeric = Numeric(0)
        the_list_empty = list_of_all_betas_in_expression(expr_numeric)
        self.assertFalse(the_list_empty)

    def test_list_of_random_variables(self):
        rv1 = RandomVariable('test')
        rv2 = RandomVariable('test2')
        dr1 = Draws('test', draw_type='NORMAL')
        dr2 = Draws('test2', draw_type='UNIFORM')
        expression = rv1 * rv2 + dr1 * dr2
        the_list = list_of_random_variables_in_expression(expression)
        the_expected_list = [rv1, rv2]
        self.assertCountEqual(
            {str(x) for x in the_list},
            {str(x) for x in the_expected_list},
        )

    def test_list_of_draws(self):
        rv1 = RandomVariable('test')
        rv2 = RandomVariable('test2')
        dr1 = Draws('test', draw_type='NORMAL')
        dr2 = Draws('test2', draw_type='UNIFORM')
        expression = rv1 * rv2 + dr1 * dr2
        the_list = list_of_draws_in_expression(expression)
        the_expected_list = [dr1, dr2]
        self.assertCountEqual(
            {str(x) for x in the_list},
            {str(x) for x in the_expected_list},
        )

    def test_getClassName(self):
        expr2 = 2 * self.beta1 * self.Variable1 - exp(-self.beta2 * self.Variable2) / (
            self.beta3 * (self.beta2 >= self.beta1)
        )
        c = expr2.get_class_name()
        self.assertEqual(c, 'Minus')

    def test_embedExpression(self):
        expr2 = 2 * self.beta1 * self.Variable1 - exp(-self.beta2 * self.Variable2) / (
            self.beta3 * (self.beta2 >= self.beta1)
        )
        self.assertTrue(expr2.embed_expression(Minus))
        self.assertFalse(expr2.embed_expression(Draws))

    def test_panel_variables(self):
        expr_ok = PanelLikelihoodTrajectory(self.Variable1)
        audit_tuple: AuditTuple = audit_expression(expr_ok)
        self.assertFalse(audit_tuple.warnings)
        self.assertFalse(audit_tuple.errors)

    def test_check_draws(self):
        d1 = Draws('d1', 'UNIFORM')
        d2 = Draws('d2', 'UNIFORM')
        expr_ok = MonteCarlo(d1 * d2)
        audit_tuple: AuditTuple = audit_expression(expr_ok)
        self.assertFalse(audit_tuple.warnings)
        self.assertFalse(audit_tuple.errors)
        expr_not_ok = d1 + MonteCarlo(d1 * d2)
        audit_tuple: AuditTuple = audit_expression(expr_not_ok)
        self.assertTrue(audit_tuple.errors)

    def test_check_rv(self):
        d1 = RandomVariable('d1')
        d2 = RandomVariable('d2')
        expr_ok = IntegrateNormal(d1, d1)
        audit_tuple: AuditTuple = audit_expression(expr_ok)
        self.assertFalse(audit_tuple.warnings)
        self.assertFalse(audit_tuple.errors)
        expr_warning = IntegrateNormal(1, d1)
        audit_tuple: AuditTuple = audit_expression(expr_warning)
        self.assertTrue(audit_tuple.warnings)
        self.assertFalse(audit_tuple.errors)
        expr_not_ok = d1 + IntegrateNormal(1, d2)
        audit_tuple: AuditTuple = audit_expression(expr_not_ok)
        self.assertTrue(audit_tuple.errors)

    def test_ids_multiple_formulas(self):
        expr1 = 2 * self.beta1 - exp(-self.beta2) / (
            self.beta2 * (self.beta3 >= self.beta4)
            + self.beta1 * (self.beta3 < self.beta4)
        )
        expr2 = 2 * self.beta1 * self.Variable1 - exp(-self.beta2 * self.Variable2) / (
            self.beta2 * (self.beta3 >= self.beta4)
            + self.beta1 * (self.beta3 < self.beta4)
        )
        collection_of_formulas = [expr1, expr2]
        registry = ExpressionRegistry(
            expressions=collection_of_formulas,
            database=self.myData,
        )
        expected_free_betas_indices = {'beta1': 0, 'beta2': 1}
        self.assertDictEqual(registry.free_betas_indices, expected_free_betas_indices)
        expected_fixed_betas_indices = {'beta3': 0, 'beta4': 1}
        self.assertDictEqual(registry.fixed_betas_indices, expected_fixed_betas_indices)
        expected_variables_indices = {
            'Av1': 5,
            'Av2': 6,
            'Av3': 7,
            'Choice': 4,
            'Exclude': 1,
            'Person': 0,
            'Variable1': 2,
            'Variable2': 3,
        }
        self.assertDictEqual(registry.variables_indices, expected_variables_indices)
        expected_draws_indices = {}
        self.assertDictEqual(registry.draws_indices, expected_draws_indices)
        expected_random_variables_indices = {}
        self.assertDictEqual(
            registry.random_variables_indices, expected_random_variables_indices
        )

    def test_expr3(self):
        my_draws = Draws('myDraws', 'UNIFORM')
        expr3 = MonteCarlo(my_draws * my_draws)
        res = get_value_c(
            expression=expr3,
            database=self.myData,
            number_of_draws=100000,
            numerically_safe=False,
        )
        for v in res:
            self.assertAlmostEqual(v, 1.0 / 3.0, 2)

    def test_expr4(self):
        omega = RandomVariable('omega')
        density = normalpdf(omega)
        a = 0
        b = 1
        x = a + (b - a) / (1 + exp(-omega))
        dx = (b - a) * exp(-omega) * (1 + exp(-omega)) ** (-2)
        integrand = x * x
        expr4 = IntegrateNormal((integrand * dx / (b - a)) / density, 'omega')
        res = get_value_c(
            expression=expr4, database=self.myData, numerically_safe=False
        )
        for v in res:
            self.assertAlmostEqual(v, 1.0 / 3.0, 2)

    def test_expr5(self):
        expr1 = 2 * self.beta1 - exp(-self.beta2) / (
            self.beta3 * (self.beta2 >= self.beta1)
        )
        expr2 = 2 * self.beta1 * self.Variable1 - exp(-self.beta2 * self.Variable2) / (
            self.beta3 * (self.beta2 >= self.beta1)
        )
        expr5 = Elem({1: expr1, 2: expr2}, self.Person) / 10
        res = list(
            get_value_c(expression=expr5, database=self.myData, numerically_safe=False)
        )
        res_ok = [
            -0.02703200460356393,
            -0.02703200460356393,
            -0.02703200460356393,
            1.6,
            2.0,
        ]
        for check_left, check_right in zip(res, res_ok):
            self.assertAlmostEqual(check_left, check_right, 5)

    def test_expr6(self):
        expr1 = 2 * self.beta1 - exp(-self.beta2) / (
            self.beta3 * (self.beta2 >= self.beta1)
        )
        expr2 = 2 * self.beta1 * self.Variable1 - exp(-self.beta2 * self.Variable2) / (
            self.beta3 * (self.beta2 >= self.beta1)
        )
        omega = RandomVariable('omega')
        density = normalpdf(omega)
        a = 0
        b = 1
        x = a + (b - a) / (1 + exp(-omega))
        dx = (b - a) * exp(-omega) * (1 + exp(-omega)) ** (-2)
        integrand = x * x
        expr4 = IntegrateNormal(integrand * dx / (b - a) / density, 'omega')
        expr6 = MultipleSum([expr1, expr2, expr4])
        res = list(
            get_value_c(expression=expr6, database=self.myData, numerically_safe=False)
        )
        res_ok = [
            4.063012266030643,
            8.063012266030643,
            12.063012266030643,
            16.063012266030643,
            20.063012266030643,
        ]
        for check_left, check_right in zip(res, res_ok):
            self.assertAlmostEqual(check_left, check_right, 3)

    def test_expr7(self):
        self.beta1.status = 1
        self.beta2.status = 1
        V = {0: -self.beta1, 1: -self.beta2, 2: -self.beta1}
        av = {0: 1, 1: 1, 2: 1}
        expr7 = LogLogit(V, av, 1)
        r = expr7.get_value()
        self.assertAlmostEqual(r, -1.2362866960692134, 5)
        expr8 = models.loglogit(V, av, 1)
        res = get_value_c(
            expression=expr8, database=self.myData, numerically_safe=False
        )
        for v in res:
            self.assertAlmostEqual(v, -1.2362866960692136, 5)

    def test_expr9(self):
        V = {0: -self.beta1, 1: -self.beta2, 2: -self.beta1}
        av = {0: 1, 1: 1, 2: 1}
        expr8 = models.loglogit(V, av, 1)
        expr9 = Derive(expr8, 'beta2')
        res = get_value_c(
            expression=expr9, database=self.myData, numerically_safe=False
        )
        for v in res:
            self.assertAlmostEqual(v, -0.7095392129298093, 5)

    def test_expr10(self):
        expr10 = NormalCdf(self.Variable1 / 10 - 1)
        res = get_value_c(
            expression=expr10, database=self.myData, numerically_safe=False
        )
        for i, j in zip(
            res,
            [
                0.5,
                0.8413447460685283,
                0.9772498680518218,
                0.99865010196837,
                0.9999683287581669,
            ],
        ):
            self.assertAlmostEqual(i, j, 5)

    def test_expr11(self):
        expr1 = 2 * self.beta1 - exp(-self.beta2) / (
            self.beta3 * (self.beta2 >= self.beta1)
        )
        expr2 = 2 * self.beta1 * self.Variable1 - exp(-self.beta2 * self.Variable2) / (
            self.beta3 * (self.beta2 >= self.beta1)
        )
        expr5 = Elem({1: expr1, 2: expr2}, self.Person) / 10
        expr10 = NormalCdf(self.Variable1 / 10 - 1)
        expr11 = BinaryMin(expr5, expr10)
        res = get_value_c(
            expression=expr11, database=self.myData, numerically_safe=False
        )
        res_ok = [-0.027032, -0.027032, -0.027032, 0.9986501, 0.99996833]
        for i, j in zip(
            res,
            res_ok,
        ):
            self.assertAlmostEqual(i, j, 5)

    def test_expr12(self):
        expr1 = 2 * self.beta1 - exp(-self.beta2) / (
            self.beta3 * (self.beta2 >= self.beta1)
        )
        expr2 = 2 * self.beta1 * self.Variable1 - exp(-self.beta2 * self.Variable2) / (
            self.beta3 * (self.beta2 >= self.beta1)
        )
        expr5 = Elem({1: expr1, 2: expr2}, self.Person) / 10
        expr10 = NormalCdf(self.Variable1 / 10 - 1)
        expr12 = BinaryMax(expr5, expr10)
        res = get_value_c(
            expression=expr12, database=self.myData, numerically_safe=False
        )
        for i, j in zip(res, [0.5, 0.8413447460685283, 0.9772498680518218, 1.6, 2.0]):
            self.assertAlmostEqual(i, j, 5)

    def test_linear_utility(self):
        terms = [
            LinearTermTuple(beta=self.beta1, x=Variable('Variable1')),
            LinearTermTuple(beta=self.beta2, x=Variable('Variable2')),
        ]
        the_utility = LinearUtility(terms)
        expression_function = create_function_simple_expression(
            expression=the_utility, database=self.myData, numerically_safe=False
        )
        the_function_output: FunctionOutput = expression_function(
            [0, 0], gradient=False, hessian=False, bhhh=False
        )
        self.assertAlmostEqual(the_function_output.function, 0, 3)
        optimization_function = NegativeLikelihood(
            dimension=2,
            loglikelihood=expression_function,
        )
        check_results = optimization_function.check_derivatives(x=[10, 11])
        for a_grad, fd_grad in zip(
            check_results.analytical.gradient, check_results.finite_differences.gradient
        ):
            self.assertAlmostEqual(a_grad, fd_grad, delta=1.0e-4)
        for a_hess, fd_hess in zip(
            np.nditer(check_results.analytical.hessian),
            np.nditer(check_results.finite_differences.hessian),
        ):
            self.assertAlmostEqual(a_hess, fd_hess, delta=1.0e-4)

    def test_expr13(self):
        newvar = self.myData.define_variable('newvar', self.Variable1 + self.Variable2)
        terms = [
            LinearTermTuple(beta=self.beta1, x=Variable('Variable1')),
            LinearTermTuple(beta=self.beta2, x=Variable('Variable2')),
            LinearTermTuple(beta=self.beta3, x=newvar),
        ]
        expr13 = LinearUtility(terms)
        res = get_value_c(
            expression=expr13, database=self.myData, numerically_safe=False
        )
        res_ok = [
            152,
            304,
            456,
            608,
            760,
        ]
        for i, j in zip(res, res_ok):
            self.assertAlmostEqual(i, j, 5)
        expr13bis = (
            self.beta1 * Variable('Variable1')
            + self.beta2 * Variable('Variable2')
            + self.beta3 * newvar
        )
        res = get_value_c(
            expression=expr13bis, database=self.myData, numerically_safe=False
        )
        res_ok = [
            152,
            304,
            456,
            608,
            760,
        ]
        for i, j in zip(res, res_ok):
            self.assertAlmostEqual(i, j, 5)

    def test_expr14(self):
        c1 = Draws('draws1', 'NORMAL_HALTON2')
        c2 = Draws('draws2', 'NORMAL_HALTON2')
        U1 = Beta('beta1', 0, None, None, 0) * Variable('Variable1') + 10 * c1
        U2 = Beta('beta2', 0, None, None, 0) * Variable('Variable2') + 10 * c2
        U3 = 0
        U = {1: U1, 2: U2, 3: U3}
        av = {1: self.Av1, 2: self.Av2, 3: self.Av3}
        expr14 = log(
            MonteCarlo(PanelLikelihoodTrajectory(models.logit(U, av, self.Choice)))
        )
        self.myData.panel('Person')
        res = get_value_c(
            expression=expr14,
            database=self.myData,
            number_of_draws=100000,
            numerically_safe=False,
        )
        res_ok = [-3.93304323, -2.10368902]
        for i, j in zip(res, res_ok):
            self.assertAlmostEqual(i, j, 3)
        # f_list, g_list, h_list, b_list
        the_function_output: FunctionOutput = get_value_and_derivatives(
            expression=expr14,
            database=self.myData,
            number_of_draws=10_000_000,
            gradient=True,
            hessian=True,
            bhhh=True,
            numerically_safe=False,
        )
        expected_function = -6.0367212197720335
        expected_gradient = [-15.76597701, 142.34022993]
        expected_hessian = [
            [-1154.32134784, 11400.42932874],
            [11400.42932874, -115432.13478364],
        ]
        expected_bhhh = [
            [168.48075656, -1150.18334338],
            [-1150.18334338, 10148.25747682],
        ]
        self.assertAlmostEqual(
            the_function_output.function, expected_function, places=3
        )
        for actual, expected in zip(the_function_output.gradient, expected_gradient):
            self.assertAlmostEqual(actual, expected, places=3)
        for actual_row, expected_row in zip(
            the_function_output.hessian, expected_hessian
        ):
            for actual, expected in zip(actual_row, expected_row):
                self.assertAlmostEqual(actual, expected, places=3)
        for actual_row, expected_row in zip(the_function_output.bhhh, expected_bhhh):
            for actual, expected in zip(actual_row, expected_row):
                self.assertAlmostEqual(actual, expected, places=3)

    def test_belongs_to(self):
        yes = BelongsTo(Numeric(2), {1, 2, 3})
        self.assertEqual(get_value_c(expression=yes, numerically_safe=False), 1.0)
        no = BelongsTo(Numeric(4), {1, 2, 3})
        self.assertEqual(get_value_c(expression=no, numerically_safe=False), 0.0)

    def test_conditional_sum(self):
        the_terms = [
            ConditionalTermTuple(condition=Numeric(0), term=Numeric(10)),
            ConditionalTermTuple(condition=Numeric(1), term=Numeric(20)),
        ]
        the_sum = ConditionalSum(the_terms)
        self.assertEqual(get_value_c(expression=the_sum, numerically_safe=False), 20)

    def test_sin(self):
        beta = Beta('Beta', 0.11, None, None, 0)
        the_sin = sin(2 * beta)
        expression_function = create_function_simple_expression(
            expression=the_sin, numerically_safe=False
        )
        the_function_output: FunctionOutput = expression_function(
            [1.0], gradient=False, hessian=False, bhhh=False
        )
        self.assertAlmostEqual(the_function_output.function, 0.909297, 3)
        optimization_function = NegativeLikelihood(
            dimension=1,
            loglikelihood=expression_function,
        )
        check_results = optimization_function.check_derivatives(x=[1.0])
        for a_grad, fd_grad in zip(
            check_results.analytical.gradient, check_results.finite_differences.gradient
        ):
            self.assertAlmostEqual(a_grad, fd_grad, places=2)
        for a_hess, fd_hess in zip(
            np.nditer(check_results.analytical.hessian),
            np.nditer(check_results.finite_differences.hessian),
        ):
            self.assertAlmostEqual(a_hess, fd_hess, places=2)

    def test_cos(self):
        beta = Beta('Beta', 0.11, None, None, 0)
        the_cos = cos(2 * beta)
        expression_function = create_function_simple_expression(
            expression=the_cos, numerically_safe=False
        )
        the_function_output: FunctionOutput = expression_function(
            [1.0], gradient=False, hessian=False, bhhh=False
        )
        self.assertAlmostEqual(the_function_output.function, -0.4161468365471424, 3)
        optimization_function = NegativeLikelihood(
            dimension=1,
            loglikelihood=expression_function,
        )
        check_results = optimization_function.check_derivatives(x=[1.0])
        for a_grad, fd_grad in zip(
            check_results.analytical.gradient, check_results.finite_differences.gradient
        ):
            self.assertAlmostEqual(a_grad, fd_grad, places=2)
        for a_hess, fd_hess in zip(
            np.nditer(check_results.analytical.hessian),
            np.nditer(check_results.finite_differences.hessian),
        ):
            self.assertAlmostEqual(a_hess, fd_hess, places=2)

    def test_bioMin(self):
        expr = BinaryMin(Numeric(3), Numeric(2))
        result = expr.get_value()
        expected_result = 2
        self.assertEqual(expected_result, result)

    def test_bioMax(self):
        expr = BinaryMax(Numeric(3), Numeric(2))
        result = expr.get_value()
        expected_result = 3
        self.assertEqual(expected_result, result)

    def test_and_2(self):
        true_expr_1: Expression = Numeric(3) == Numeric(3)
        true_expr_2: Expression = Numeric(2) == Numeric(2)
        false_expr: Expression = Numeric(3) == Numeric(2)
        result = true_expr_1 & true_expr_2
        self.assertEqual(result.get_value(), 1)
        result = true_expr_1 & false_expr
        self.assertEqual(result.get_value(), 0)

    def test_and_3(self):
        true_expr: Expression = Numeric(3) == Numeric(3)
        false_expr_1: Expression = Numeric(3) == Numeric(2)
        false_expr_2: Expression = Numeric(2) == Numeric(3)
        result: Expression = true_expr | false_expr_1
        self.assertEqual(result.get_value(), 1)
        result = false_expr_1 | false_expr_2
        self.assertEqual(result.get_value(), 0)


class TestExpressionConversion(unittest.TestCase):
    def setUp(self):
        self.beta = Beta('beta', 42, None, None, 0)

    def test_validate_and_convert_numeric(self):
        self.assertIsInstance(validate_and_convert(3.14), Numeric)

    def test_validate_and_convert_boolean(self):
        self.assertEqual(validate_and_convert(True).value, Numeric(1).value)
        self.assertEqual(validate_and_convert(False).value, Numeric(0).value)

    def test_validate_and_convert_invalid_type(self):
        with self.assertRaises(TypeError):
            validate_and_convert("invalid")

    def test_expression_to_value_numeric(self):
        self.assertEqual(expression_to_value(3.14), 3.14)

    def test_expression_to_value_boolean(self):
        self.assertEqual(expression_to_value(True), 1.0)

    def test_expression_to_value_invalid_type(self):
        with self.assertRaises(TypeError):
            expression_to_value("invalid")

    def test_expression_to_value_expression(self):
        self.assertEqual(expression_to_value(self.beta), 42.0)

    def test_get_dict_values(self):
        test_dict = {1: 3.14, 2: True, 3: self.beta}
        expected = {1: 3.14, 2: 1.0, 3: 42.0}
        self.assertEqual(get_dict_values(test_dict), expected)

    def test_get_dict_expressions(self):
        test_dict = {1: 3.14, 2: True, 3: self.beta}
        result = get_dict_expressions(test_dict)
        self.assertIsInstance(result[1], Numeric)
        self.assertIsInstance(result[2], Numeric)
        self.assertIsInstance(result[3], Beta)


class TestConvertToDict(unittest.TestCase):
    def test_basic_conversion(self):
        sequence = ['apple', 'banana', 'cherry']
        name_to_index = {'first': 0, 'second': 1, 'third': 2}
        expected_result = {'first': 'apple', 'second': 'banana', 'third': 'cherry'}
        self.assertEqual(convert_to_dict(sequence, name_to_index), expected_result)

    def test_empty_sequence(self):
        sequence = []
        name_to_index = {}
        expected_result = {}
        self.assertEqual(convert_to_dict(sequence, name_to_index), expected_result)

    def test_empty_map(self):
        sequence = ['apple', 'banana', 'cherry']
        name_to_index = {}
        expected_result = {}
        self.assertEqual(convert_to_dict(sequence, name_to_index), expected_result)

    def test_index_out_of_bounds(self):
        sequence = ['apple', 'banana']
        name_to_index = {'first': 0, 'third': 2}
        with self.assertRaises(IndexError):
            convert_to_dict(sequence, name_to_index)

    def test_negative_index(self):
        sequence = ['apple', 'banana', 'cherry']
        name_to_index = {'first': -1}
        with self.assertRaises(IndexError):
            convert_to_dict(sequence, name_to_index)

    def test_large_index_in_small_list(self):
        sequence = ['apple']
        name_to_index = {'second': 1}
        with self.assertRaises(IndexError):
            convert_to_dict(sequence, name_to_index)

    def test_with_non_integer_indices(self):
        sequence = ['apple', 'banana', 'cherry']
        name_to_index = {'first': '0'}  # index as a string by mistake
        with self.assertRaises(TypeError):  # Assumes TypeError for non-integers
            convert_to_dict(sequence, name_to_index)


if __name__ == '__main__':
    unittest.main()



import pandas as pd

from biogeme.exceptions import BiogemeError
from biogeme.deprecated import deprecated


def count_number_of_groups(df: pd.DataFrame, column: str) -> int:
    """
    This function counts the number of groups of same value in a column.
    For instance: 1,2,2,3,3,3,4,1,1  would give 5.

    Example::

        >>>df = pd.DataFrame({'ID': [1, 1, 2, 3, 3, 1, 2, 3],
                              'value':[1000,
                                       2000,
                                       3000,
                                       4000,
                                       5000,
                                       5000,
                                       10000,
                                       20000]})
        >>>count_number_of_groups(df,'ID')
        6

        >>>count_number_of_groups(df,'value')
        7

    """
    df['_bio_groups'] = pd.Series(df[column] != df[column].shift(1)).cumsum()
    result = len(df['_bio_groups'].unique())
    df.drop(columns=['_bio_groups'], inplace=True)
    return result


@deprecated(count_number_of_groups)
def countNumberOfGroups(df: pd.DataFrame, column: str) -> int:
    pass


def flatten_database(
    df: pd.DataFrame,
    merge_id: str,
    row_name: str | None = None,
    identical_columns: list[str] | None = None,
) -> pd.DataFrame:
    """Combine several rows of a Pandas database into one.
    For instance, consider the following database::

           ID  Age  Cost   Name
        0   1   23    34  Item3
        1   1   23    45  Item4
        2   1   23    12  Item7
        3   2   45    65  Item3
        4   2   45    34  Item7



    If row_name is 'Name', the function generates the same data in the
    following format::

            Age  Item3_Cost  Item4_Cost  Item7_Cost
        ID
        1    23          34        45.0          12
        2    45          65         NaN          34


    If row_name is None, the function generates the same data in the
    following format::

            Age  1_Cost 1_Name  2_Cost 2_Name  3_Cost 3_Name
        ID
        1    23      34  Item3      45  Item4    12.0  Item7
        2    45      65  Item3      34  Item7     NaN    NaN

    :param df: initial data frame
    :type df: pandas.DataFrame

    :param merge_id: name of the column that identifies rows that
        should be merged. In the above example: 'ID'
    :type merge_id: str

    :param row_name: name of the columns that provides the name of the
        rows in the new dataframe. In the example above: 'Name'. If
        None, the rows are numbered sequentially.
    :type row_name: str

    :param identical_columns: name of the columns that contain
        identical values across the rows of a group. In the example
        above: ['Age']. If None, these columns are automatically
        detected. On large database, there may be a performance issue.
    :type identical_columns: list(str)

    :return: reformatted database
    :rtype: pandas.DataFrame
    """
    df_copy = df.copy()
    all_columns = set(df_copy.columns)
    duplicate = f'{merge_id}_biogeme_tmp_duplicate'
    df_copy[duplicate] = df_copy.loc[:, merge_id]
    grouped = df_copy.groupby(by=duplicate)

    def are_values_identical(col: pd.Series) -> bool:
        """This function checks if all the values in a column
        are identical

        :param col: the column

        :return: True if all values are identical. False otherwise.
        """

        return (col.iloc[0] == col).all(0)

    def get_varying_cols(g: pd.DataFrame) -> set[str]:
        """This functions returns the name of all columns
        that have constant values within each group of data.

        :param g: group of data

        :return: name of all columns that have constant values
            within each group of data.
        """
        return {colname for colname, col in g.items() if not are_values_identical(col)}

    if identical_columns is None:
        all_varying_cols = grouped.apply(get_varying_cols, include_groups=False)
        varying_columns = set.union(*all_varying_cols)
        identical_columns = list(all_columns - varying_columns)
        varying_columns = list(varying_columns)
    else:
        identical_columns = set(identical_columns)
        identical_columns.add(merge_id)
        varying_columns = list(all_columns - identical_columns)

    # Take the first row for columns that are identical
    if identical_columns:
        common_data = df_copy[list(identical_columns)].drop_duplicates(
            merge_id, keep='first'
        )
        common_data.index = common_data[merge_id]
    # Treat the other columns
    # Include merge_id and a duplicate
    tmp_df = df_copy[[merge_id] + list(varying_columns)].copy()
    tmp_df[duplicate] = tmp_df[merge_id].copy()
    grouped_varying = tmp_df.groupby(by=duplicate)

    def treat(x: pd.DataFrame) -> pd.DataFrame:
        """Treat a group of data.

        :param x: group of data

        :return: the same data organized in one row, with proper column names

        :raise BiogemeError:  if there are duplicates in the name of
        the row. Indeed, in that case, they cannot be used to name the
        new columns.
        """
        if not are_values_identical(x[merge_id]):
            err_msg = f'Group has different IDs: {x[merge_id]}. ' f'Rows id: {x.index}'
            raise BiogemeError(err_msg)
        if row_name is not None and not x[row_name].is_unique:
            err_msg = (
                f'Entries in column [{row_name}] are not unique. '
                f'This column cannot be used to name the new '
                f'columns:\n{x[[row_name, merge_id]]}. '
            )
            raise BiogemeError(err_msg)

        the_columns = set(x.columns) - {merge_id}
        if row_name is not None:
            the_columns -= {row_name}
        sorted_list = sorted(list(the_columns))
        first = True
        i = 0
        for _, row in x.iterrows():
            i += 1
            if first:
                all_values = [row[merge_id]]
                all_columns = [merge_id]
                first = False
            name = f'{i}' if row_name is None else row[row_name]
            columns = [f'{name}_{c}' for c in sorted_list]
            all_values.extend([row[c] for c in sorted_list])
            all_columns.extend(columns)
        treated_df = pd.DataFrame([all_values], columns=all_columns)
        return treated_df

    flat_data = grouped_varying.apply(treat, include_groups=False)
    flat_data.index = flat_data[merge_id]

    # We remove the column 'merge_id' as it is stored as index.
    if identical_columns:
        return pd.concat([common_data, flat_data], axis='columns').drop(
            columns=[merge_id]
        )
    return flat_data.drop(columns=[merge_id])



"""

Mixture of logit with panel data
================================

Example of a mixture of logit models, using Monte-Carlo integration.
 The datafile is organized as panel data.

:author: Michel Bierlaire, EPFL
:date: Sun Apr  9 18:12:17 2023

"""

import numpy as np
import biogeme.biogeme_logging as blog
import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import (
    Beta,
    bioDraws,
    PanelLikelihoodTrajectory,
    MonteCarlo,
    log,
)

# %%
# See the data processing script: :ref:`swissmetro_panel`.
from swissmetro_panel import (
    database,
    CHOICE,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
)

logger = blog.get_screen_logger(level=blog.INFO)
logger.info('Example b12panel.py')

# %%
# We set the seed so that the results are reproducible. This is not necessary in general.
np.random.seed(seed=90267)

# %%
# Parameters to be estimated.
B_COST = Beta('B_COST', 0, None, None, 0)

# %%
# Define a random parameter, normally distributed across individuals,
# designed to be used for Monte-Carlo simulation.
B_TIME = Beta('B_TIME', 0, None, None, 0)

# %%
# It is advised not to use 0 as starting value for the following parameter.
B_TIME_S = Beta('B_TIME_S', 1, None, None, 0)
B_TIME_RND = B_TIME + B_TIME_S * bioDraws('B_TIME_RND', 'NORMAL_ANTI')

# %%
# We do the same for the constants, to address serial correlation.
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_CAR_S = Beta('ASC_CAR_S', 1, None, None, 0)
ASC_CAR_RND = ASC_CAR + ASC_CAR_S * bioDraws('ASC_CAR_RND', 'NORMAL_ANTI')

ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_TRAIN_S = Beta('ASC_TRAIN_S', 1, None, None, 0)
ASC_TRAIN_RND = ASC_TRAIN + ASC_TRAIN_S * bioDraws('ASC_TRAIN_RND', 'NORMAL_ANTI')

ASC_SM = Beta('ASC_SM', 0, None, None, 1)
ASC_SM_S = Beta('ASC_SM_S', 1, None, None, 0)
ASC_SM_RND = ASC_SM + ASC_SM_S * bioDraws('ASC_SM_RND', 'NORMAL_ANTI')

# %%
# Definition of the utility functions.
V1 = ASC_TRAIN_RND + B_TIME_RND * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM_RND + B_TIME_RND * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR_RND + B_TIME_RND * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# %%
# Associate utility functions with the numbering of alternatives.
V = {1: V1, 2: V2, 3: V3}

# %%
# Associate the availability conditions with the alternatives.
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# %%
# Conditional on the random parameters, the likelihood of one observation is
# given by the logit model (called the kernel).
obsprob = models.logit(V, av, CHOICE)

# %%
# Conditional on the random parameters, the likelihood of all observations for
# one individual (the trajectory) is the product of the likelihood of
# each observation.
condprobIndiv = PanelLikelihoodTrajectory(obsprob)

# %%
# We integrate over the random parameters using Monte-Carlo
logprob = log(MonteCarlo(condprobIndiv))

# %%
# Create the Biogeme object. As the objective is to illustrate the
# syntax, we calculate the Monte-Carlo approximation with a small
# number of draws. To achieve that, we provide a parameter file
# different from the default one: `<few_draws.toml>`_
the_biogeme = bio.BIOGEME(database, logprob, parameter_file='few_draws.toml')
the_biogeme.modelName = 'b12panel'

# %%
# Estimate the parameters.
results = the_biogeme.estimate()

# %%
print(results.short_summary())

# %%
pandas_results = results.getEstimatedParameters()
pandas_results



"""Arithmetic expressions accepted by Biogeme: binary division

Michel Bierlaire
Sat Jun 14 2025, 10:23:07
"""

from __future__ import annotations

import logging

import jax.numpy as jnp
from biogeme.exceptions import BiogemeError
from biogeme.floating_point import EPSILON

from .base_expressions import ExpressionOrNumeric
from .binary_expressions import BinaryOperator
from .jax_utils import JaxFunctionType
from .numeric_expressions import Numeric

logger = logging.getLogger(__name__)


class Divide(BinaryOperator):
    """
    Division expression
    """

    def __init__(self, left: ExpressionOrNumeric, right: ExpressionOrNumeric):
        """Constructor

        :param left: first arithmetic expression
        :type left: biogeme.expressions.Expression

        :param right: second arithmetic expression
        :type right: biogeme.expressions.Expression
        """
        super().__init__(left, right)
        self.simplified = None
        if isinstance(self.left, Numeric) and self.left.get_value() == 0:
            self.simplified = Numeric(0)
        elif isinstance(self.right, Numeric) and self.right.get_value() == 1:
            self.simplified = self.left

    def deep_flat_copy(self) -> Divide:
        """Provides a copy of the expression. It is deep in the sense that it generates copies of the children.
        It is flat in the sense that any `MultipleExpression` is transformed into the currently selected expression.
        The flat part is irrelevant for this expression.
        """
        left_copy = self.left.deep_flat_copy()
        right_copy = self.right.deep_flat_copy()
        return type(self)(left=left_copy, right=right_copy)

    def __str__(self) -> str:
        return f'({self.left} / {self.right})'

    def __repr__(self) -> str:
        return f'({repr(self.left)} / {repr(self.right)})'

    def get_value(self) -> float:
        """Evaluates the value of the expression

        :return: value of the expression
        :rtype: float
        """
        if self.simplified is not None:
            return self.simplified.get_value()
        left_val = self.left.get_value()
        right_val = self.right.get_value()
        if right_val == 0.0:
            raise BiogemeError("Division by zero in Divide expression.")
        return left_val / right_val

    def recursive_construct_jax_function(
        self, numerically_safe: bool
    ) -> JaxFunctionType:
        """
        Generates a function to be used by biogeme_jax. Must be overloaded by each expression
        :return: the function takes two parameters: the parameters, and one row of the database.
        """
        if self.simplified is not None:
            return self.simplified.recursive_construct_jax_function(
                numerically_safe=numerically_safe
            )
        left_jax: JaxFunctionType = self.left.recursive_construct_jax_function(
            numerically_safe=numerically_safe
        )
        right_jax: JaxFunctionType = self.right.recursive_construct_jax_function(
            numerically_safe=numerically_safe
        )
        if numerically_safe:

            def the_jax_function(
                parameters: jnp.ndarray,
                one_row: jnp.ndarray,
                the_draws: jnp.ndarray,
                the_random_variables: jnp.ndarray,
            ) -> float:
                left_value = left_jax(
                    parameters, one_row, the_draws, the_random_variables
                )
                right_value = right_jax(
                    parameters, one_row, the_draws, the_random_variables
                )
                return left_value * right_value / (right_value**2 + EPSILON)

            return the_jax_function

        def the_jax_function(
            parameters: jnp.ndarray,
            one_row: jnp.ndarray,
            the_draws: jnp.ndarray,
            the_random_variables: jnp.ndarray,
        ) -> float:
            left_value = left_jax(parameters, one_row, the_draws, the_random_variables)
            right_value = right_jax(
                parameters, one_row, the_draws, the_random_variables
            )

            return left_value / right_value

        return the_jax_function



import unittest

import pandas as pd

from biogeme.database import Database
from biogeme.expressions import Beta, Variable
from biogeme.expressions_registry import ExpressionRegistry


class TestExpressionRegistry(unittest.TestCase):
    def setUp(self):
        # Set up a simple Biogeme model
        self.data = pd.DataFrame(
            {
                'x': [1.0, 2.0, 3.0],
                'y': [4.0, 5.0, 6.0],
            }
        )
        self.database = Database('test', self.data)

        self.beta1 = Beta('beta1', 1.0, -10.0, 10.0, 0)
        self.beta2 = Beta('beta2', 2.0, 0.0, 5.0, 0)
        self.x = Variable('x')
        self.y = Variable('y')

        self.expr = self.beta1 * self.x + self.beta2 * self.y
        self.registry = ExpressionRegistry([self.expr], self.database)

    def test_expressions_property(self):
        self.assertEqual(len(self.registry.expressions), 1)
        self.assertIs(self.registry.expressions[0], self.expr)

    def test_free_betas_extraction(self):
        free_betas = self.registry.free_betas
        free_betas_names = {beta.name for beta in free_betas}
        self.assertIn('beta1', free_betas_names)
        self.assertIn('beta2', free_betas_names)

    def test_fixed_betas_empty(self):
        self.assertEqual(len(self.registry.fixed_betas), 0)

    def test_free_betas_indices(self):
        indices = self.registry.free_betas_indices
        self.assertIn('beta1', indices)
        self.assertIn('beta2', indices)
        self.assertEqual(sorted(indices.values()), [0, 1])

    def test_bounds(self):
        bounds = self.registry.bounds
        self.assertEqual(bounds, [(-10.0, 10.0), (0.0, 5.0)])

    def test_free_betas_values(self):
        values = self.registry.free_betas_init_values
        self.assertEqual(values['beta1'], 1.0)
        self.assertEqual(values['beta2'], 2.0)

    def test_requires_draws_false(self):
        self.assertFalse(self.registry.requires_draws)

    def test_draws_empty(self):
        self.assertEqual(len(self.registry.draws), 0)

    def test_variables_detected(self):
        vars_ = self.registry.variables
        vars_names = [var.name for var in vars_]
        self.assertIn('x', vars_names)
        self.assertIn('y', vars_names)

    def test_variables_indices(self):
        indices = self.registry.variables_indices
        self.assertIn('x', indices)
        self.assertIn('y', indices)

    def test_number_of_free_betas(self):
        self.assertEqual(self.registry.number_of_free_betas, 2)

    def test_draw_types_empty(self):
        self.assertEqual(self.registry.draw_types(), {})

    def test_broadcast_assigns_ids(self):
        # Check that beta1 has a specific_id set
        self.registry.broadcast()
        self.assertIsNotNone(self.registry.free_betas[0].specific_id)
        self.assertIsInstance(self.registry.free_betas[0].specific_id, int)


if __name__ == '__main__':
    unittest.main()



import unittest

import biogeme.biogeme as bio
from biogeme import models
from biogeme.data.swissmetro import (
    CAR_AV_SP,
    CAR_CO_SCALED,
    CAR_TT_SCALED,
    CHOICE,
    GA,
    PURPOSE,
    SM_AV,
    SM_CO,
    SM_COST_SCALED,
    SM_TT_SCALED,
    TRAIN_AV_SP,
    TRAIN_CO,
    TRAIN_COST_SCALED,
    TRAIN_TT_SCALED,
    read_data,
)
from biogeme.expressions import Beta, Draws, MonteCarlo, exp, log
from biogeme.parameters import Parameters

database = read_data()
# Keep only trip purposes 1 (commuter) and 3 (business)
exclude = ((PURPOSE != 1) * (PURPOSE != 3)) > 0
database.remove(exclude)

ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_TIME_S = Beta('B_TIME_S', 1, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)

# Define a random parameter, normally distirbuted, designed to be used
# for Monte-Carlo simulation
B_TIME_RND = B_TIME + B_TIME_S * Draws('b_time_rnd', 'UNIFORMSYM')

# Utility functions

# If the person has a GA (season ticket) her incremental cost is actually 0
# rather than the cost value gathered from the
# network data.
SM_COST = SM_CO * (GA == 0)
TRAIN_COST = TRAIN_CO * (GA == 0)


V1 = ASC_TRAIN + B_TIME_RND * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + B_TIME_RND * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME_RND * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}


av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# The choice model is a logit, with availability conditions
prob = exp(models.loglogit(V, av, CHOICE))
log_prob = log(MonteCarlo(prob))


class test_06(unittest.TestCase):
    def testEstimation(self):
        parameters = Parameters()
        parameters.set_value(section='MonteCarlo', name='number_of_draws', value=5)
        parameters.set_value(section='MonteCarlo', name='seed', value=10)
        biogeme = bio.BIOGEME(
            database,
            log_prob,
            parameters=parameters,
            save_iterations=False,
            generate_html=False,
            generate_yaml=False,
        )
        results = biogeme.estimate()
        self.assertEqual(results.number_of_parameters, 5)


if __name__ == '__main__':
    unittest.main()



"""Implements objects in charge of the management of nests for the
nested and the cross-nested logit model.

:author: Michel Bierlaire
:date: Thu Oct  5 14:45:58 2023

"""

from __future__ import annotations

import itertools
import logging
from dataclasses import dataclass
from typing import Iterator, Any

import numpy as np
import pandas as pd
from scipy.integrate import dblquad

from biogeme.exceptions import BiogemeError
from biogeme.expressions import (
    Expression,
    Numeric,
    ExpressionOrNumeric,
    Beta,
    get_dict_expressions,
    get_dict_values,
)


# The first versions of Biogeme defined the nests as follows. We keep his for backward compatibility.
OldOneNestForNestedLogit = tuple[ExpressionOrNumeric, list[int]]
OldNestsForNestedLogit = tuple[OldOneNestForNestedLogit, ...]
OldOneNestForCrossNestedLogit = tuple[ExpressionOrNumeric, dict[int, Expression]]
OldNestsForCrossNestedLogit = tuple[OldOneNestForCrossNestedLogit, ...]

logger = logging.getLogger(__name__)


@dataclass
class OneNestForNestedLogit:
    """Class capturing the information for one nest of the nested logit model"""

    nest_param: ExpressionOrNumeric
    list_of_alternatives: list[int]
    name: str | None = None

    @classmethod
    def from_tuple(cls, the_tuple: OldOneNestForNestedLogit) -> OneNestForNestedLogit:
        """Ctor to initialize the nest using the old syntax of Biogeme with a tuple"""
        return cls(*the_tuple)

    def intersection(self, other_nest: OneNestForNestedLogit) -> set[int]:
        """Returns the intersection of two nests. Designed to verify the
        validity of the specification

        """
        set1 = set(self.list_of_alternatives)
        set2 = set(other_nest.list_of_alternatives)
        return set1 & set2


def get_nest(a_nest: OneNestForNestedLogit | OldOneNestForNestedLogit):
    """Convert a nest definition to an OneNestForNestedLogot, if needed."""

    if isinstance(a_nest, OneNestForNestedLogit):
        return a_nest
    if isinstance(a_nest, tuple):
        return OneNestForNestedLogit.from_tuple(a_nest)
    raise TypeError(f'Object to type {type(a_nest)} does not represent a nest.')


@dataclass
class OneNestForCrossNestedLogit:
    """Tuple capturing the information for one nest of the cross-nested logit"""

    nest_param: ExpressionOrNumeric
    dict_of_alpha: dict[int, ExpressionOrNumeric]

    name: str | None = None

    @classmethod
    def from_tuple(
        cls, the_tuple: OldOneNestForCrossNestedLogit
    ) -> OneNestForCrossNestedLogit:
        """Ctor to initialize the nest using the old syntax of Biogeme with a tuple"""
        return cls(*the_tuple)

    def __post_init__(self) -> None:
        self.dict_of_alpha = get_dict_expressions(self.dict_of_alpha)
        self.list_of_alternatives = list(self.dict_of_alpha.keys())

    def all_alpha_fixed(self) -> bool:
        """Check if the alpha parameters have a numeric value."""
        for _, alpha in self.dict_of_alpha.items():
            if isinstance(alpha, Beta):
                if alpha.status == 0:
                    return False
            elif isinstance(alpha, Expression) and not isinstance(alpha, Numeric):
                return False
        return True


class Nests:
    """Generic interface for the nests."""

    def __init__(
        self,
        choice_set: list[int],
        tuple_of_nests: tuple[OneNestForNestedLogit | OneNestForCrossNestedLogit, ...],
    ):
        """Ctor

        :param choice_set: the list of all alternatives in the choice
            set. We use a list instead of a set because the order
            matters.

        :param tuple_of_nests: the list of nests

        """

        self.choice_set = choice_set
        self.tuple_of_nests = tuple_of_nests

        # Provide names to nests if there is none.
        nest_counter = 0
        for nest in self.tuple_of_nests:
            nest_counter += 1
            if nest.name is None:
                nest.name = f'nest_{nest_counter}'

        # Set of alternatives involved in nests
        self.mev_alternatives = set().union(
            *(set(nest.list_of_alternatives) for nest in self.tuple_of_nests)
        )
        # Verify that all elements are indeed in the choice set
        invalid_elements = self.mev_alternatives - set(self.choice_set)
        if invalid_elements:
            raise BiogemeError(
                f'The following alternatives appear in a nest and not in the '
                f'choice set: {invalid_elements}'
            )

        self.alone = set(self.choice_set) - self.mev_alternatives
        if self.alone:
            warning_msg = (
                f'The following elements do not appear in any nest and are assumed each to be alone in a separate '
                f'nest: {self.alone}. If it is not the intention, check the assignment of alternatives to nests.'
            )
            logger.warning(warning_msg)

    def __getitem__(
        self, index: int
    ) -> OneNestForNestedLogit | OneNestForCrossNestedLogit:
        if index < 0 or index >= len(self.tuple_of_nests):
            raise IndexError(
                f'Index out of bounds. Valid indices are between 0 and '
                f'{len(self.tuple_of_nests) - 1}.'
            )
        return self.tuple_of_nests[index]

    def __iter__(self) -> Iterator[Any]:
        return iter(self.tuple_of_nests)

    def check_names(self) -> bool:
        """Checks that all the nests have a name"""
        for nest in self.tuple_of_nests:
            if nest.name is None:
                return False
        return True

    def check_union(self) -> tuple[bool, str]:
        """Check if the union of the nests is the choice set

        :return: a boolean with the result of the check, as a message if check fails.
        """

        # Union Check: The union of all lists should be equal to choice_set

        union_of_lists = {
            i for nest in self.tuple_of_nests for i in nest.list_of_alternatives
        }
        union_of_lists |= set(self.alone)
        if union_of_lists != set(self.choice_set):
            missing_values = set(self.choice_set) - union_of_lists  # set difference
            extra_values = union_of_lists - set(
                self.choice_set
            )  # values not in choice_set
            error_msg_1 = (
                f'Alternatives in the choice set, '
                f'but not in any nest: {missing_values}.'
                if missing_values
                else ''
            )
            error_msg_2 = (
                f'Alternatives in a nest, but not in the choice set: {extra_values}.'
                if extra_values
                else ''
            )
            error_msg = f'{error_msg_1} {error_msg_2}'
            logger.error(error_msg)
            return False, error_msg
        return True, ''


class NestsForNestedLogit(Nests):
    """This class handles nests for the nested logit model"""

    def __init__(
        self,
        choice_set: list[int],
        tuple_of_nests: tuple[OneNestForNestedLogit, ...] | OldNestsForNestedLogit,
    ):
        """Ctor

        :param choice_set: the list of all alternatives in the choice set
        :param tuple_of_nests: the list of nests. The old syntax can still be used:
            A tuple containing as many items as nests.
            Each item is also a tuple containing two items:

            - an object of type biogeme.expressions.expr.Expression representing
              the nest parameter,
            - a list containing the list of identifiers of the alternatives
              belonging to the nest.

            Example::

                nesta = MUA ,[1, 2, 3]
                nestb = MUB ,[4, 5, 6]
                nests = nesta, nestb


        """

        # In previous versions of Biogeme, the nests were defined
        # using regular tuples, not NamedTuple. We cast them here for
        # the sake of backward compatibility.
        if not all(isinstance(elem, OneNestForNestedLogit) for elem in tuple_of_nests):
            tuple_of_nests = tuple(
                OneNestForNestedLogit.from_tuple(nest) for nest in tuple_of_nests
            )
        super().__init__(choice_set, tuple_of_nests)

    def correlation(
        self,
        parameters: dict[str, float] | None = None,
        alternatives_names: dict[int, str] | None = None,
        mu: float = 1.0,
    ) -> pd.DataFrame:
        """Calculate the correlation matrix of the error terms of all
            alternatives of a nested logit model.

        :param parameters: values of the parameters.

        :param alternatives_names: dictionary associating a name with each alternative.

        :param mu: value of the scale parameter mu.

        :return: correlation matrix
        """
        index = {alt: i for i, alt in enumerate(self.choice_set)}
        nbr_of_alternatives = len(self.choice_set)
        if alternatives_names is None:
            alternatives_names = {i: str(i) for i in self.choice_set}
        correlation = np.identity(nbr_of_alternatives)
        for m in self.tuple_of_nests:
            if isinstance(m.nest_param, Expression):
                if parameters:
                    m.nest_param.change_init_values(parameters)
                mu_m = m.nest_param.get_value()
            else:
                mu_m = m.nest_param
            alt_m = m.list_of_alternatives
            for i, j in itertools.combinations(alt_m, 2):
                correlation[index[i]][index[j]] = correlation[index[j]][index[i]] = (
                    1.0 - 1.0 / (mu_m * mu_m)
                    if mu == 1.0
                    else 1.0 - (mu * mu) / (mu_m * mu_m)
                )
        return pd.DataFrame(
            correlation,
            index=list(alternatives_names.values()),
            columns=list(alternatives_names.values()),
        )

    def check_intersection(self) -> tuple[bool, str]:
        """Check if the intersection of nests is empty

        :return: a boolean with the result of the check, and a message if check fails.

        """
        for i, nest in enumerate(self.tuple_of_nests):
            if set(nest.list_of_alternatives) & self.alone:
                error_msg = (
                    f'The following alternatives are both in nest {nest.name} '
                    f'and identified as alone in a nest: '
                    f'{set(nest.list_of_alternatives) & self.alone}'
                )
                logger.error(error_msg)
                return False, error_msg
            for j, other_nest in enumerate(self.tuple_of_nests):
                if i != j:
                    the_intersection = nest.intersection(other_nest)
                    if the_intersection:
                        error_msg = (
                            f'The following alternatives appear both in nests '
                            f'{nest.name} and {other_nest.name}: {the_intersection}.'
                        )
                        logger.error(error_msg)
                        return False, error_msg
        return True, ''

    def check_partition(self) -> tuple[bool, str]:
        """Check if the nests correspond to a partition

        :return: a boolean with the result of the check, and a message if check fails.
        """
        valid_union, msg_union = self.check_union()
        valid_intersection, msg_intersection = self.check_intersection()
        return valid_union and valid_intersection, msg_union + '; ' + msg_intersection


class NestsForCrossNestedLogit(Nests):
    """This class handles nests for the cross-nested logit model"""

    def __init__(
        self,
        choice_set: list[int],
        tuple_of_nests: (
            tuple[OneNestForCrossNestedLogit, ...] | OldNestsForCrossNestedLogit
        ),
    ):
        """Ctor

        :param choice_set: the list of all alternatives in the choice set
        :param tuple_of_nests: the list of nests

        """

        # In previous versions of Biogeme, the nests were defined
        # using regular tuples, not NamedTuple. We cast them here for
        # the sake of backward compatibility.
        if not all(
            isinstance(elem, OneNestForCrossNestedLogit) for elem in tuple_of_nests
        ):
            tuple_of_nests = tuple(
                OneNestForCrossNestedLogit.from_tuple(nest) for nest in tuple_of_nests
            )
        super().__init__(choice_set, tuple_of_nests)

    def all_alphas_fixed(self) -> bool:
        """Check if all the alphas are fixed"""
        for nest in self.tuple_of_nests:
            if not nest.all_alpha_fixed():
                return False
        return True

    def get_alpha_dict(self, alternative_id: int) -> dict[str, Expression]:
        """Generates a dict mapping each nest with the alpha
        parameter, for a given alternative

        :param alternative_id: identifier of the alternative
        :return: a dict mapping the name of a nest and the alpha expression
        """
        alphas = {
            nest.name: nest.dict_of_alpha.get(alternative_id, Numeric(0.0))
            for nest in self.tuple_of_nests
        }
        return alphas

    def get_alpha_values(self, alternative_id: int) -> dict[str, float]:
        """Generates a dict mapping each nest with the value of the alpha
        parameters, for a given alternative

        :param alternative_id: identifier of the alternative
        :return: a dict mapping the name of a nest and the value of the alpha expression
        """

        alpha_dict = self.get_alpha_dict(alternative_id=alternative_id)
        alpha_values = {
            key: expression.get_value() for key, expression in alpha_dict.items()
        }
        return alpha_values

    def check_validity(self) -> tuple[bool, str]:
        """Verifies if the cross-nested logit specification is valid

        :return: a boolean with the result of the check, and a message if check fails.
        """
        ok, message = self.check_union()

        alt: dict[int, list[Expression]] = {i: [] for i in self.choice_set}
        number = 0
        for nest in self.tuple_of_nests:
            alpha = nest.dict_of_alpha
            for i, a in alpha.items():
                if a.get_value() != 0.0:
                    alt[i].append(a)
            number += 1

        problems_one = []
        for i, ell in alt.items():
            if len(ell) == 1 and isinstance(ell[0], Expression):
                problems_one.append(i)

        if problems_one:
            message += (
                f' Alternative in exactly one nest, '
                f'and parameter alpha is defined by an '
                f'expression, and may not be constant: {problems_one}'
            )

        return ok, message

    def covariance(
        self, i: int, j: int, parameters: dict[str, float] | None = None
    ) -> float:
        """Calculate the covariance between the error terms of two
        alternatives of a cross-nested logit model. It is assumed that
        the homogeneity parameter mu of the model has been normalized
        to one.

        :param parameters: values of the parameters.

        :param i: first alternative

        :param j: second alternative

        :return: value of the correlation

        :raise BiogemeError: if the requested number is non-positive or a float

        """

        if i not in self.choice_set:
            raise BiogemeError(f'Unknown alternative: {i}')
        if j not in self.choice_set:
            raise BiogemeError(f'Unknown alternative: {j}')

        if i == j:
            return np.pi * np.pi / 6.0

        def integrand(z_i: float, z_j: float) -> float:
            """Function to be integrated to calculate the correlation between
            alternative i and alternative j.

            :param z_i: argument corresponding to alternative i
            :type z_i: float

            :param z_j: argument corresponding to alternative j
            :type z_j: float
            """
            y_i = -np.log(z_i)
            y_j = -np.log(z_j)
            xi_i = -np.log(y_i)
            xi_j = -np.log(y_j)
            dy_i = -1.0 / z_i
            dy_j = -1.0 / z_j
            dxi_i = -dy_i / y_i
            dxi_j = -dy_j / y_j

            g_sum = 0.0
            gi_sum = 0.0
            gj_sum = 0.0
            gij_sum = 0.0
            for m in self.tuple_of_nests:
                if isinstance(m.nest_param, Expression):
                    if parameters:
                        m.nest_param.change_init_values(parameters)
                    mu_m = m.nest_param.get_value()
                else:
                    mu_m = m.nest_param
                alphas = get_dict_values(m.dict_of_alpha, parameters)
                alpha_i = alphas.get(i, 0)
                if alpha_i != 0:
                    term_i = (alpha_i * y_i) ** mu_m
                else:
                    term_i = 0
                alpha_j = alphas.get(j, 0)
                if alpha_j != 0:
                    term_j = (alpha_j * y_j) ** mu_m
                else:
                    term_j = 0
                the_sum = term_i + term_j
                p1 = (1.0 / mu_m) - 1
                p2 = (1.0 / mu_m) - 2
                g_sum += the_sum ** (1.0 / mu_m)
                if alpha_i != 0:
                    gi_sum += alpha_i**mu_m * y_i ** (mu_m - 1) * the_sum**p1
                if alpha_j != 0:
                    gj_sum += alpha_j**mu_m * y_j ** (mu_m - 1) * the_sum**p1
                if mu_m != 1.0 and alpha_i != 0 and alpha_j != 0:
                    gij_sum += (
                        (1 - mu_m)
                        * the_sum**p2
                        * (alpha_i * alpha_j) ** mu_m
                        * (y_i * y_j) ** (mu_m - 1)
                    )

            f = np.exp(-g_sum)
            f_second = f * y_i * y_j * (gi_sum * gj_sum - gij_sum)

            return xi_i * xi_j * f_second * dxi_i * dxi_j

        integral, _ = dblquad(integrand, 0, 1, lambda x: 0, lambda x: 1)
        return integral - np.euler_gamma * np.euler_gamma

    def correlation(
        self,
        parameters: dict[str, float] | None = None,
        alternatives_names: dict[int, str] | None = None,
    ) -> pd.DataFrame:
        """Calculate the correlation matrix of the error terms of all
            alternatives of cross-nested logit model.

        :param parameters: values of the parameters.
        :param alternatives_names: names of the alternative, for
            better reporting. If not provided, the number are used.
        :return: correlation matrix

        """
        nbr_of_alternatives = len(self.choice_set)
        if alternatives_names is None:
            alternatives_names = {i: str(i) for i in self.choice_set}

        covar = np.empty((nbr_of_alternatives, nbr_of_alternatives))
        for i, alt_i in enumerate(self.choice_set):
            for j, alt_j in enumerate(self.choice_set):
                covar[i][j] = self.covariance(alt_i, alt_j, parameters)
                if i != j:
                    covar[j][i] = covar[i][j]

        v = np.sqrt(np.diag(covar))
        outer_v = np.outer(v, v)
        correlation = covar / outer_v
        correlation[covar == 0] = 0
        return pd.DataFrame(
            correlation,
            index=list(alternatives_names.values()),
            columns=list(alternatives_names.values()),
        )



"""

biogeme.biogeme
===============

Examples of use of several functions.

This is designed for programmers who need examples of use of the
functions of the module. The examples are designed to illustrate the
syntax. They do not correspond to any meaningful model.

:author: Michel Bierlaire
:date: Thu Nov 16 18:36:35 2023
"""

import biogeme.version as ver
import biogeme.biogeme as bio
import biogeme.database as db
import pandas as pd
from biogeme.expressions import Beta, Variable, exp
import biogeme.biogeme_logging as blog

# %%
# Version of Biogeme.
print(ver.getText())


# %%
# Logger.
logger = blog.get_screen_logger(level=blog.INFO)
logger.info('Logger initalized')


# %%
# Definition of a database
df = pd.DataFrame(
    {
        'Person': [1, 1, 1, 2, 2],
        'Exclude': [0, 0, 1, 0, 1],
        'Variable1': [1, 2, 3, 4, 5],
        'Variable2': [10, 20, 30, 40, 50],
        'Choice': [1, 2, 3, 1, 2],
        'Av1': [0, 1, 1, 1, 1],
        'Av2': [1, 1, 1, 1, 1],
        'Av3': [0, 1, 1, 1, 1],
    }
)
myData = db.Database('test', df)

# %%
# Data
myData.data


# %%
# Definition of various expressions.
Variable1 = Variable('Variable1')
Variable2 = Variable('Variable2')
beta1 = Beta('beta1', -1.0, -3, 3, 0)
beta2 = Beta('beta2', 2.0, -3, 10, 0)
likelihood = -(beta1**2) * Variable1 - exp(beta2 * beta1) * Variable2 - beta2**4
simul = beta1 / Variable1 + beta2 / Variable2
dictOfExpressions = {'loglike': likelihood, 'beta1': beta1, 'simul': simul}

# %%
# Creation of the BIOGEME object.
myBiogeme = bio.BIOGEME(myData, dictOfExpressions)
myBiogeme.modelName = 'simple_example'
print(myBiogeme)

# %%
# The data is stored in the Biogeme object.
myBiogeme.database.data

# %%
# Log likelihood with the initial values of the parameters.
myBiogeme.calculateInitLikelihood()

# %%
# Calculate the log-likelihood with a different value of the
# parameters. We retieve the current value and add 1 to each of them.
x = myBiogeme.id_manager.free_betas_values
xplus = [v + 1 for v in x]
print(xplus)

# %%
myBiogeme.calculateLikelihood(xplus, scaled=True)

# %%
# Calculate the log-likelihood function and its derivatives.
f, g, h, bhhh = myBiogeme.calculateLikelihoodAndDerivatives(
    xplus, scaled=True, hessian=True, bhhh=True
)
# %%
print(f'f = {f}')

# %%
print(f'g = {g}')

# %%
pd.DataFrame(h)

# %%
pd.DataFrame(bhhh)

# %%
# Now the unscaled version.
f, g, h, bhhh = myBiogeme.calculateLikelihoodAndDerivatives(
    xplus, scaled=False, hessian=True, bhhh=True
)
# %%
print(f'f = {f}')

# %%
print(f'g = {g}')

# %%
pd.DataFrame(h)

# %%
pd.DataFrame(bhhh)


# %%
# Calculate the hessian of the log likelihood function using finite difference.
fin_diff_hessian = myBiogeme.likelihoodFiniteDifferenceHessian(xplus)
pd.DataFrame(fin_diff_hessian)

# %%
# Check numerically the derivatives implementation. The analytical
# derivatives are compared to the numerical derivatives obtains by
# finite differences.
f, g, h, gdiff, hdiff = myBiogeme.checkDerivatives(xplus, verbose=True)

# %%
print(f'f = {f}')
# %%
print(f'g = {g}')
# %%
pd.DataFrame(h)
# %%
pd.DataFrame(gdiff)
# print(f'gdiff = {gdiff}')
# %%
pd.DataFrame(hdiff)
# print(f'hdiff = {hdiff}')

# %%
# Estimation
# ----------

# %%
# Estimation of the parameters, with bootstrapping
myBiogeme.bootstrap_samples = 10
results = myBiogeme.estimate(run_bootstrap=True)

# %%
results.getEstimatedParameters()

# %%
# If the model has already been estimated, it is possible to recycle
# the estimation results. In that case, the other arguments are
# ignored, and the results are whatever is in the file.

# %%
recycled_results = myBiogeme.estimate(recycle=True, run_bootstrap=True)

# %%
print(recycled_results.short_summary())

# %%
recycled_results.getEstimatedParameters()

# %%
# Simulation
# ----------

# %%
# Simulate with the initial values for the parameters.
simulation_with_default_betas = myBiogeme.simulate(myBiogeme.loglike.get_beta_values())
simulation_with_default_betas

# %%
# Simulate with the estimated values for the parameters.

# %%
print(results.getBetaValues())

# %%
simulation_with_estimated_betas = myBiogeme.simulate(results.getBetaValues())
simulation_with_estimated_betas

# %%
# Confidence intervals.
# First, we extract the values of betas from the bootstrapping draws.
draws_from_betas = results.getBetasForSensitivityAnalysis(
    myBiogeme.id_manager.free_betas.names
)
for draw in draws_from_betas:
    print(draw)

# %%
# Then, we calculate the confidence intervals. The default interval
# size is 0.9. Here, we use a different one.
left, right = myBiogeme.confidenceIntervals(draws_from_betas, interval_size=0.95)
left

# %%
right

# %%
# Validation
# ----------
# The validation consists in organizing the data into several slices
# of about the same size, randomly defined.  Each slide is considered
# as a validation dataset. The model is then re-estimated using all
# the data except the slice, and the estimated model is applied on the
# validation set (i.e. the slice). The value of the log likelihood for
# each observation in the validation set is reported in a
# dataframe. As this is done for each slice, the output is a list of
# dataframes, each corresponding to one of these exercises.

# %%
validationData = myData.split(slices=5)
validation_results = myBiogeme.validate(results, validationData)

# %%
for slide in validation_results:
    print(
        f'Log likelihood for {slide.shape[0]} '
        f'validation data: {slide["Loglikelihood"].sum()}'
    )


# %%
# The following tools is used to find files with the model name and a
# specific extension.
myBiogeme.files_of_type('pickle')



"""Function to visit the tree of Expressions

Michel Bierlaire
15.04.2025 12:55
"""

from typing import Any

from .base_expressions import Expression


class ExpressionVisitor:
    """Walks the tree and executes side-effect handlers with no return values."""

    def __init__(self):
        self._registry = {}

    def register(self, expr_type: type[Expression]):
        """
        Register a handler function for a specific expression type. The handler must
        return None and perform side effects only.

        :param expr_type: the type of Expression for which the handler should be used
        :return: decorator that registers the handler
        """

        def decorator(func):
            self._registry[expr_type] = func
            return func

        return decorator

    def visit(self, expr: Expression, context: Any = None) -> None:
        """
        Traverse the expression tree and apply handlers for side effects.

        :param expr: the root expression to visit
        :param context: optional context object passed to handlers
        """
        if context is None:
            context = {}
        if 'ancestors' not in context:
            context['ancestors'] = []
        context['ancestors'].append(expr)

        handler = self._registry.get(type(expr))
        if handler:
            result = handler(expr, context)
            if result is not None:
                raise TypeError(
                    f"Handler for {type(expr).__name__} must return None, got {type(result).__name__}"
                )
        for child in expr.get_children():
            self.visit(child, context)

        context['ancestors'].pop()



"""Generation of the default parameter file and values

:author: Michel Bierlaire
:date: Wed Nov 30 10:14:35 2022

IMPORTANT: when only one "check" function is provided, do not forget
to insert a comma at the end, before the closing parenthesis for the
tuple.
See https://www.w3schools.com/python/gloss_python_tuple_one_item.asp
"""

from __future__ import annotations

from typing import Callable, NamedTuple

import biogeme.check_parameters as cp
import biogeme.optimization as opt
import numpy as np
from biogeme.floating_point import NUMPY_FLOAT
from biogeme.second_derivatives import SecondDerivativesMode
from biogeme.version import get_version
from tabulate import tabulate

ParameterValue = bool | int | float | str

MISSING_VALUE = 99999


class ParameterTuple(NamedTuple):
    name: str
    value: ParameterValue
    type: type
    section: str
    description: str
    check: tuple[Callable[[ParameterValue], tuple[bool, str | None]], ...]


def all_parameters_tuple() -> tuple[ParameterTuple, ...]:
    return (
        ParameterTuple(
            name='identification_threshold',
            value=1.0e-5,
            type=float,
            section='Output',
            description=(
                'float: if the smallest eigenvalue of the second derivative '
                'matrix is lesser or equal to this parameter, the model is '
                'considered not identified. The corresponding eigenvector '
                'is then reported to identify the parameters involved in the issue.'
            ),
            check=(cp.is_number,),
        ),
        ParameterTuple(
            name='only_robust_stats',
            value=True,
            type=bool,
            section='Output',
            description=(
                'bool: "True" if only the robust statistics need to be reported.'
                ' If "False", the statistics from the Rao-Cramer bound are '
                'also reported.'
            ),
            check=(cp.is_boolean,),
        ),
        ParameterTuple(
            name='generate_html',
            value=True,
            type=bool,
            section='Output',
            description=(
                'bool: "True" if the HTML file with the results must be generated.'
            ),
            check=(cp.is_boolean,),
        ),
        ParameterTuple(
            name='generate_yaml',
            value=True,
            type=bool,
            section='Output',
            description=(
                'bool: "True" if the yaml file with the results must be generated.'
            ),
            check=(cp.is_boolean,),
        ),
        ParameterTuple(
            name='save_validation_results',
            value=True,
            type=bool,
            section='Output',
            description=(
                'bool: "True" if the validation results are saved in CSV files.'
            ),
            check=(cp.is_boolean,),
        ),
        ParameterTuple(
            name='number_of_threads',
            value=0,
            type=int,
            section='MultiThreading',
            description=(
                'int: Number of threads/processors to be used. If'
                ' the parameter is 0, the number of'
                ' available threads is calculated using'
                ' cpu_count().'
            ),
            check=(cp.is_integer, cp.is_non_negative),
        ),
        ParameterTuple(
            name='number_of_draws',
            value=10_000,
            type=int,
            section='MonteCarlo',
            description='int: Number of draws for Monte-Carlo integration.',
            check=(cp.is_integer, cp.is_positive),
        ),
        ParameterTuple(
            name='missing_data',
            value=MISSING_VALUE,
            type=int,
            section='Specification',
            description=(
                'number: If one variable has this value, it is assumed'
                ' that a data is missing and an exception will'
                ' be triggered.'
            ),
            check=(cp.is_number,),
        ),
        ParameterTuple(
            name='numerically_safe',
            value=False,
            type=bool,
            section='Specification',
            description=(
                'If true, Biogeme is doing its best to deal with numerical issues, '
                'such as division by a number close to zero, at the possible expense of speed.'
            ),
            check=(cp.is_boolean,),
        ),
        ParameterTuple(
            name='seed',
            value=0,
            type=int,
            section='MonteCarlo',
            description=(
                'int: Seed used for the pseudo-random number generation. It is'
                ' useful only when each run should generate the exact same'
                ' result. If 0, a new seed is used at each run.'
            ),
            check=(cp.is_integer, cp.is_non_negative),
        ),
        ParameterTuple(
            name='bootstrap_samples',
            value=100,
            type=int,
            section='Estimation',
            description='int: number of re-estimations for bootstrap sampling.',
            check=(cp.is_integer, cp.is_non_negative),
        ),
        ParameterTuple(
            name='calculating_second_derivatives',
            value='analytical',
            type=str,
            section='Estimation',
            description=(
                f'Defines how to calculate the second derivatives: {",".join([value.value for value in SecondDerivativesMode])}. '
            ),
            check=(cp.check_calculating_second_derivatives,),
        ),
        ParameterTuple(
            name='large_data_set',
            value=100_000,
            type=int,
            section='Estimation',
            description=(
                'If the number of observations is larger than this value, the data set is deemed large, and the '
                'default estimation algorithm will not use second derivatives.'
            ),
            check=(cp.is_integer, cp.is_non_negative),
        ),
        ParameterTuple(
            name='max_number_parameters_to_report',
            value=15,
            type=int,
            section='Estimation',
            description=(
                'int: maximum number of parameters to report during the estimation.'
            ),
            check=(cp.is_integer, cp.is_non_negative),
        ),
        ParameterTuple(
            name='save_iterations',
            value=True,
            type=bool,
            section='Estimation',
            description=(
                'bool: If True, the current iterate is saved after'
                ' each iteration, in a file named'
                ' ``__[modelName].iter``, where'
                ' ``[modelName]`` is the name given to the'
                ' model. If such a file exists, the starting'
                ' values for the estimation are replaced by'
                ' the values saved in the file.'
            ),
            check=(cp.is_boolean,),
        ),
        ParameterTuple(
            name='maximum_number_catalog_expressions',
            value=100,
            type=int,
            section='Estimation',
            description=(
                'If the expression contains catalogs, the parameter sets an '
                'upper bound of the total number of possible combinations '
                'that can be estimated in the same loop.'
            ),
            check=(
                cp.is_integer,
                cp.is_positive,
            ),
        ),
        ParameterTuple(
            name='optimization_algorithm',
            value='automatic',
            type=str,
            section='Estimation',
            description=(
                f'str: optimization algorithm to be used for estimation. '
                f'Valid values: {["automatic"]+list(opt.algorithms.keys())}'
            ),
            check=(cp.check_algo_name,),
        ),
        ParameterTuple(
            name='second_derivatives',
            value=1.0,
            type=float,
            section='SimpleBounds',
            description=(
                'float: proportion (between 0 and 1) of iterations when the '
                'analytical Hessian is calculated'
            ),
            check=(cp.zero_one, cp.is_number),
        ),
        ParameterTuple(
            name='tolerance',
            value=float(np.finfo(NUMPY_FLOAT).eps ** (1.0 / 3.0)),
            type=float,
            section='SimpleBounds',
            description='float: the algorithm stops when this precision is reached',
            check=(cp.is_number,),
        ),
        ParameterTuple(
            name='max_iterations',
            value=1000,
            type=int,
            section='SimpleBounds',
            description='int: maximum number of iterations',
            check=(cp.is_integer, cp.is_positive),
        ),
        ParameterTuple(
            name='infeasible_cg',
            value=False,
            type=bool,
            section='SimpleBounds',
            description=(
                'If True, the conjugate gradient algorithm may generate '
                'infeasible solutions until termination.  The result will '
                'then be projected on the feasible domain.  If False, the '
                'algorithm stops as soon as an infeasible iterate is '
                'generated'
            ),
            check=(cp.is_boolean,),
        ),
        ParameterTuple(
            name='initial_radius',
            value=1,
            type=float,
            section='SimpleBounds',
            description='Initial radius of the trust region',
            check=(cp.is_number, cp.is_positive),
        ),
        ParameterTuple(
            name='steptol',
            value=float(np.finfo(NUMPY_FLOAT).eps ** (2.0 / 3.0)),
            type=float,
            section='SimpleBounds',
            description=(
                'The algorithm stops when the relative change in x is below '
                'this threshold. Basically, if p significant digits of x '
                'are needed, steptol should be set to 1.0e-p.'
            ),
            check=(cp.is_number, cp.is_positive),
        ),
        ParameterTuple(
            name='enlarging_factor',
            value=10,
            type=float,
            section='SimpleBounds',
            description=(
                'If an iteration is very successful, the radius of '
                'the trust region is multiplied by this factor'
            ),
            check=(cp.is_number, cp.is_positive),
        ),
        ParameterTuple(
            name='dogleg',
            value=True,
            type=bool,
            section='TrustRegion',
            description=(
                'bool: choice of the method to solve the trust region subproblem. '
                'True: dogleg. False: truncated conjugate gradient.'
            ),
            check=(cp.is_boolean,),
        ),
        ParameterTuple(
            name='maximum_number_parameters',
            value=50,
            type=int,
            section='AssistedSpecification',
            description=(
                'int: maximum number of parameters allowed in a model. Each specification '
                'with a higher number is deemed invalid and not estimated.'
            ),
            check=(cp.is_integer, cp.is_positive),
        ),
        ParameterTuple(
            name='number_of_neighbors',
            value=20,
            type=int,
            section='AssistedSpecification',
            description=(
                'int: maximum number of neighbors that are visited by the VNS algorithm.'
            ),
            check=(cp.is_integer, cp.is_non_negative),
        ),
        ParameterTuple(
            name='largest_neighborhood',
            value=20,
            type=int,
            section='AssistedSpecification',
            description=(
                'int: size of the largest neighborhood copnsidered by the Variable '
                'Neighborhood Search (VNS) algorithm.'
            ),
            check=(cp.is_integer, cp.is_non_negative),
        ),
        ParameterTuple(
            name='maximum_attempts',
            value=100,
            type=int,
            section='AssistedSpecification',
            description=(
                'int: an attempts consists in selecting a solution in the Pareto '
                'set, and trying to improve it. The parameter imposes an upper bound '
                'on the total number of attempts, irrespectively if they are '
                'successful or not.'
            ),
            check=(cp.is_integer, cp.is_non_negative),
        ),
        ParameterTuple(
            name='number_of_jobs',
            value=-1,
            type=int,
            section='Bootstrap',
            description=(
                'int: The maximum number of concurrently running jobs. If -1 is given, joblib tries to use all CPUs.'
            ),
            check=(cp.is_integer,),
        ),
        ParameterTuple(
            name='version',
            value=get_version(),
            type=str,
            section='Biogeme',
            description='Version of Biogeme that created the TOML file. Do not modify this value.',
            check=(),
        ),
    )


def print_list_of_parameters(table_format='plain') -> str:
    """Generate a table describing all the parameters"""

    headers = ['Parameter', 'Default value', 'Type', 'Section', 'Description']
    rows = [
        [
            parameter.name,
            parameter.value,
            parameter.type,
            parameter.section,
            parameter.description,
        ]
        for parameter in all_parameters_tuple()
    ]
    return tabulate(rows, headers=headers, tablefmt='plain')



"""

Mixture of logit
================

Choice model with latent variable. No measurement equation for the indicators.

:author: Michel Bierlaire, EPFL
:date: Thu Apr 13 16:58:21 2023

"""

import biogeme.biogeme_logging as blog
import biogeme.biogeme as bio
from biogeme import models
import biogeme.distributions as dist
from biogeme.expressions import (
    Beta,
    RandomVariable,
    Integrate,
    exp,
    log,
)

from read_or_estimate import read_or_estimate

from optima import (
    database,
    age_65_more,
    ScaledIncome,
    moreThanOneCar,
    moreThanOneBike,
    individualHouse,
    male,
    haveChildren,
    haveGA,
    highEducation,
    WaitingTimePT,
    Choice,
    TimePT_scaled,
    TimeCar_scaled,
    MarginalCostPT_scaled,
    CostCarCHF_scaled,
    distance_km_scaled,
    PurpHWH,
    PurpOther,
)


logger = blog.get_screen_logger(level=blog.INFO)
logger.info('Example b03choice_only.py')

# %%
# Parameters to be estimated
coef_intercept = Beta('coef_intercept', 0.0, None, None, 1)
coef_age_65_more = Beta('coef_age_65_more', 0.0, None, None, 0)
coef_haveGA = Beta('coef_haveGA', 0.0, None, None, 0)
coef_moreThanOneCar = Beta('coef_moreThanOneCar', 0.0, None, None, 0)
coef_moreThanOneBike = Beta('coef_moreThanOneBike', 0.0, None, None, 0)
coef_individualHouse = Beta('coef_individualHouse', 0.0, None, None, 0)
coef_male = Beta('coef_male', 0.0, None, None, 0)
coef_haveChildren = Beta('coef_haveChildren', 0.0, None, None, 0)
coef_highEducation = Beta('coef_highEducation', 0.0, None, None, 0)

# %%
# Latent variable: structural equation.

# %%
# Define a random parameter, normally distributed)
# designed to be used
# for numerical integration.
omega = RandomVariable('omega')
density = dist.normalpdf(omega)
sigma_s = Beta('sigma_s', 1, None, None, 0)

thresholds = [None, 4, 6, 8, 10, None]
formula_income = models.piecewiseFormula(variable=ScaledIncome, thresholds=thresholds)

CARLOVERS = (
    coef_intercept
    + coef_age_65_more * age_65_more
    + formula_income
    + coef_moreThanOneCar * moreThanOneCar
    + coef_moreThanOneBike * moreThanOneBike
    + coef_individualHouse * individualHouse
    + coef_male * male
    + coef_haveChildren * haveChildren
    + coef_haveGA * haveGA
    + coef_highEducation * highEducation
    + sigma_s * omega
)

# %%
# Choice model: parameters.
ASC_CAR = Beta('ASC_CAR', 0.0, None, None, 0)
ASC_PT = Beta('ASC_PT', 0.0, None, None, 1)
ASC_SM = Beta('ASC_SM', 0.0, None, None, 0)
BETA_COST_HWH = Beta('BETA_COST_HWH', 0.0, None, None, 0)
BETA_COST_OTHER = Beta('BETA_COST_OTHER', 0.0, None, None, 0)
BETA_DIST = Beta('BETA_DIST', 0.0, None, None, 0)
BETA_TIME_CAR_REF = Beta('BETA_TIME_CAR_REF', -0.0001, None, 0, 0)
BETA_TIME_CAR_CL = Beta('BETA_TIME_CAR_CL', -1.0, -3, 3, 0)
BETA_TIME_PT_REF = Beta('BETA_TIME_PT_REF', -0.0001, None, 0, 0)
BETA_TIME_PT_CL = Beta('BETA_TIME_PT_CL', -1.0, -3, 3, 0)
BETA_WAITING_TIME = Beta('BETA_WAITING_TIME', 0.0, None, None, 0)

# %%
# Definition of utility functions.
BETA_TIME_PT = BETA_TIME_PT_REF * exp(BETA_TIME_PT_CL * CARLOVERS)

V0 = (
    ASC_PT
    + BETA_TIME_PT * TimePT_scaled
    + BETA_WAITING_TIME * WaitingTimePT
    + BETA_COST_HWH * MarginalCostPT_scaled * PurpHWH
    + BETA_COST_OTHER * MarginalCostPT_scaled * PurpOther
)

BETA_TIME_CAR = BETA_TIME_CAR_REF * exp(BETA_TIME_CAR_CL * CARLOVERS)

V1 = (
    ASC_CAR
    + BETA_TIME_CAR * TimeCar_scaled
    + BETA_COST_HWH * CostCarCHF_scaled * PurpHWH
    + BETA_COST_OTHER * CostCarCHF_scaled * PurpOther
)

V2 = ASC_SM + BETA_DIST * distance_km_scaled

# %%
# Associate utility functions with the numbering of alternatives.
V = {0: V0, 1: V1, 2: V2}

# %%
# Conditional on  omega, we have a logit model (called the kernel).
condprob = models.logit(V, None, Choice)

# %%
# We integrate over omega using numerical integration.
loglike = log(Integrate(condprob * density, 'omega'))


# %%
# Create the Biogeme object.
the_biogeme = bio.BIOGEME(database, loglike)
the_biogeme.modelName = 'b03choice_only'

# %%
# If estimation results are saved on file, we read them to speed up the process.
# If not, we estimate the parameters.
results = read_or_estimate(the_biogeme=the_biogeme, directory='saved_results')

# %%
print(f'Estimated betas: {len(results.data.betaValues)}')
print(f'Final log likelihood: {results.data.logLike:.3f}')
print(f'Output file: {results.data.htmlFileName}')

# %%
results.getEstimatedParameters()



from typing import NamedTuple

from biogeme.database import Database
from biogeme.model_elements import ModelElements
from biogeme.validation.prepare_validation import split


class EstimationValidationModels(NamedTuple):
    estimation: ModelElements
    validation: ModelElements


def split_databases(
    model_elements: ModelElements, slices: int, groups: str | None = None
) -> list[EstimationValidationModels]:
    """

    :param model_elements: modeling elements, including the database and the draws that will be split.
    :param slices: The number of folds/slices. Must be >= 2.
    :param groups: Optional name of the column containing group identifiers.
                       If provided, all rows with the same group ID are kept in the same fold.

    :return: A list of EstimationValidationIndices tuples, one per fold.
    """
    slices = split(
        dataframe=model_elements.database.dataframe, slices=slices, groups=groups
    )
    results = []
    database_name = model_elements.database.name
    for index, split_indices in enumerate(slices, 1):
        estimation_df = model_elements.database.dataframe.iloc[split_indices.estimation]
        estimation_data = Database(
            name=f'{database_name} estimation {index}',
            dataframe=estimation_df,
        )
        estimation_draws_management = model_elements.draws_management.extract_slice(
            split_indices.estimation
        )
        estimation_model_elements = ModelElements(
            expressions=model_elements.expressions,
            database=estimation_data,
            draws_management=estimation_draws_management,
        )
        validation_df = model_elements.database.dataframe.iloc[split_indices.validation]
        validation_data = Database(
            name=f'{database_name} estimation {index}',
            dataframe=validation_df,
        )
        validation_draws_management = model_elements.draws_management.extract_slice(
            split_indices.validation
        )
        validation_model_elements = ModelElements(
            expressions={
                f'{name} [validation fold {index}]': expression
                for name, expression in model_elements.expressions.items()
            },
            database=validation_data,
            draws_management=validation_draws_management,
        )
        the_pair = EstimationValidationModels(
            estimation=estimation_model_elements, validation=validation_model_elements
        )
        results.append(the_pair)
    return results



"""

Cross point elasticities
========================

We use a previously estimated nested logit model and calculate
disaggregate and aggregate cross point elasticities.

Details about this example are available in Section 3 of `Bierlaire (2018)
Calculating indicators with PandasBiogeme
<http://transp-or.epfl.ch/documents/technicalReports/Bier18a.pdf>`_

:author: Michel Bierlaire, EPFL
:date: Wed Apr 12 21:00:58 2023

"""

import sys
import biogeme.biogeme as bio
from biogeme import models
import biogeme.exceptions as excep
import biogeme.results as res
from biogeme.expressions import Derive
from optima_data import database, normalized_weight

from scenarios import (
    scenario,
    TimePT,
    TimeCar,
    MarginalCostPT,
    CostCarCHF,
)


# %%
# Obtain the specification for the default scenario
# The definition of the scenarios is available in :ref:`scenarios`.
V, nests, _, _ = scenario()

# %%
# Obtain the expression for the choice probability of each alternative.
prob_PT = models.nested(V, None, nests, 0)
prob_CAR = models.nested(V, None, nests, 1)
prob_SM = models.nested(V, None, nests, 2)

# %%
# The choice model is a nested logit.
prob_PT = models.nested(V, None, nests, 0)
prob_CAR = models.nested(V, None, nests, 1)
prob_SM = models.nested(V, None, nests, 2)

# %%
# Calculation of the cross elasticities.
# We use the 'Derive' operator to calculate the derivatives.
cross_elas_pt_time = Derive(prob_PT, 'TimeCar') * TimeCar / prob_PT
cross_elas_pt_cost = Derive(prob_PT, 'CostCarCHF') * CostCarCHF / prob_PT
cross_elas_car_time = Derive(prob_CAR, 'TimePT') * TimePT / prob_CAR
cross_elas_car_cost = Derive(prob_CAR, 'MarginalCostPT') * MarginalCostPT / prob_CAR

# %%
# Formulas to simulate.
simulate = {
    'weight': normalized_weight,
    'Prob. car': prob_CAR,
    'Prob. public transportation': prob_PT,
    'Prob. slow modes': prob_SM,
    'cross_elas_pt_time': cross_elas_pt_time,
    'cross_elas_pt_cost': cross_elas_pt_cost,
    'cross_elas_car_time': cross_elas_car_time,
    'cross_elas_car_cost': cross_elas_car_cost,
}

# %%
# Create the Biogeme object.
the_biogeme = bio.BIOGEME(database, simulate)

# %%
# Read the estimation results from the file
try:
    results = res.bioResults(pickleFile='saved_results/b02estimation.pickle')
except excep.BiogemeError:
    sys.exit(
        'Run first the script b02estimation.py in order to generate '
        'the file b02estimation.pickle.'
    )

# %%
# `simulated_values` is a Panda dataframe with the same number of rows as
# the database, and as many columns as formulas to simulate.
simulated_values = the_biogeme.simulate(results.getBetaValues())
simulated_values

# %%
# We calculate the aggregate elasticities.

# %%
# First, the weighted probabilities.
simulated_values['Weighted prob. car'] = (
    simulated_values['weight'] * simulated_values['Prob. car']
)
simulated_values['Weighted prob. PT'] = (
    simulated_values['weight'] * simulated_values['Prob. public transportation']
)

# %%
# Then the denominators of the aggregate elasticity expressions.
denominator_car = simulated_values['Weighted prob. car'].sum()
denominator_pt = simulated_values['Weighted prob. PT'].sum()

# %%
# And finally the aggregate elasticities themselves.

# %%
# Elasticity of car with respect to public transportation travel time.
cross_elas_term_car_time = (
    simulated_values['Weighted prob. car']
    * simulated_values['cross_elas_car_time']
    / denominator_car
).sum()
print(
    f'Aggregate cross elasticity of car wrt PT time: ' f'{cross_elas_term_car_time:.3g}'
)

# %%
# Elasticity of car with respect to public transportation travel cost.
cross_elas_term_car_cost = (
    simulated_values['Weighted prob. car']
    * simulated_values['cross_elas_car_cost']
    / denominator_car
).sum()
print(
    f'Aggregate cross elasticity of car wrt PT cost: ' f'{cross_elas_term_car_cost:.3g}'
)

# %%
# Elasticity of public transportatiom with respect to car travel time.
cross_elas_term_pt_time = (
    simulated_values['Weighted prob. PT']
    * simulated_values['cross_elas_pt_time']
    / denominator_pt
).sum()
print(
    f'Aggregate cross elasticity of PT wrt car time: ' f'{cross_elas_term_pt_time:.3g}'
)

# %%
# Elasticity of public transportatiom with respect to car travel cost.
cross_elas_term_pt_cost = (
    simulated_values['Weighted prob. PT']
    * simulated_values['cross_elas_pt_cost']
    / denominator_pt
).sum()
print(
    f'Aggregate cross direct elasticity of PT wrt car cost: '
    f'{cross_elas_term_pt_cost:.3g}'
)



"""

Estimation of mixtures of logit
===============================

Estimation of a mixtures of logit models where the integral is
calculated using numerical integration.

:author: Michel Bierlaire, EPFL
:date: Thu Apr 13 21:03:03 2023
"""

import biogeme.biogeme as bio
import biogeme.distributions as dist
from biogeme import models
from biogeme.expressions import Beta, RandomVariable, Integrate, log

from swissmetro import (
    database,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
    TRAIN_AV_SP,
    SM_AV,
    CAR_AV_SP,
    CHOICE,
)

# %%
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_TIME_S = Beta('B_TIME_S', 1, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)

# %%
# Define a random parameter, normally distirbuted, designed to be used
# for Monte-Carlo simulation
omega = RandomVariable('omega')
density = dist.normalpdf(omega)
B_TIME_RND = B_TIME + B_TIME_S * omega

# %%
# Definition of the utility functions
V1 = ASC_TRAIN + B_TIME_RND * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + B_TIME_RND * SM_TT_SCALED + B_COST * SM_COST_SCALED
V3 = ASC_CAR + B_TIME_RND * CAR_TT_SCALED + B_COST * CAR_CO_SCALED

# %%
# Associate utility functions with the numbering of alternatives
V = {1: V1, 2: V2, 3: V3}

# %%
# Associate the availability conditions with the alternatives
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# %%
# The choice model is a logit, with availability conditions
condprob = models.logit(V, av, CHOICE)
prob = Integrate(condprob * density, 'omega')
logprob = log(prob)

# %%
the_biogeme = bio.BIOGEME(database, logprob)
the_biogeme.modelName = '06estimationIntegral'

# %%
results = the_biogeme.estimate()

# %%
print(results.short_summary())

# %%
pandas_results = results.getEstimatedParameters()
pandas_results



import unittest

import biogeme.biogeme as bio
from biogeme import models
from biogeme.data.swissmetro import (
    CAR_AV_SP,
    CAR_CO_SCALED,
    CAR_TT_SCALED,
    CHOICE,
    GA,
    PURPOSE,
    SM_AV,
    SM_CO,
    SM_COST_SCALED,
    SM_TT_SCALED,
    TRAIN_AV_SP,
    TRAIN_CO,
    TRAIN_COST_SCALED,
    TRAIN_TT_SCALED,
    read_data,
)
from biogeme.expressions import Beta, Draws, MonteCarlo, PanelLikelihoodTrajectory, log
from biogeme.results_processing import EstimationResults

database = read_data()
# Keep only trip purposes 1 (commuter) and 3 (business)
exclude = ((PURPOSE != 1) * (PURPOSE != 3)) > 0
database.remove(exclude)
database.panel('ID')

ASC_CAR = Beta('ASC_CAR', -0.039775, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', -0.739801, None, None, 0)
ASC_SM = Beta('ASC_SM', 0, None, None, 1)
B_TIME = Beta('B_TIME', -6.131332, None, 0, 0)
B_COST = Beta('B_COST', -2.955561, None, 0, 0)

SIGMA_CAR = Beta('SIGMA_CAR', 3.830629, None, None, 0)
SIGMA_SM = Beta('SIGMA_SM', 0.936053, None, None, 0)
SIGMA_TRAIN = Beta('SIGMA_TRAIN', 2.898149, None, None, 0)

# Define a random parameter, normally distributed, designed to be used
# for Monte-Carlo simulation
EC_CAR = SIGMA_CAR * Draws('EC_CAR', 'NORMAL')
EC_SM = SIGMA_SM * Draws('EC_SM', 'NORMAL')
EC_TRAIN = SIGMA_TRAIN * Draws('EC_TRAIN', 'NORMAL')

SM_COST = SM_CO * (GA == 0)
TRAIN_COST = TRAIN_CO * (GA == 0)


# For latent class 1, where the time coefficient is zero
V11 = ASC_TRAIN + B_COST * TRAIN_COST_SCALED + EC_TRAIN
V12 = ASC_SM + B_COST * SM_COST_SCALED + EC_SM
V13 = ASC_CAR + B_COST * CAR_CO_SCALED + EC_CAR

V1 = {1: V11, 2: V12, 3: V13}

# For latent class 2, where the time coefficient is estimated
V21 = ASC_TRAIN + B_TIME * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED + EC_TRAIN
V22 = ASC_SM + B_TIME * SM_TT_SCALED + B_COST * SM_COST_SCALED + EC_SM
V23 = ASC_CAR + B_TIME * CAR_TT_SCALED + B_COST * CAR_CO_SCALED + EC_CAR

V2 = {1: V21, 2: V22, 3: V23}


av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}


# Class membership model
W_OTHER = Beta('W_OTHER', 0.803059, 0, 1, 0)
probClass1 = 1 - W_OTHER
probClass2 = W_OTHER

# The choice model is a discrete mixture of logit, with availability conditions
prob1 = PanelLikelihoodTrajectory(models.logit(V1, av, CHOICE))
prob2 = PanelLikelihoodTrajectory(models.logit(V2, av, CHOICE))
prob_indiv = probClass1 * prob1 + probClass2 * prob2
log_prob = log(MonteCarlo(prob_indiv))


class test_15(unittest.TestCase):
    def testEstimation(self):
        biogeme = bio.BIOGEME(
            database,
            log_prob,
            number_of_draws=100,
            seed=1111,
            save_iterations=False,
            generate_html=False,
            generate_yaml=False,
        )
        results: EstimationResults = biogeme.estimate()
        self.assertAlmostEqual(results.final_log_likelihood, -3639.291359844362, 2)


if __name__ == '__main__':
    unittest.main()
